""" Provides class Fitter with helping functions
"""
import logging
import numpy as np
from scipy.optimize import curve_fit

fitter_logger = logging.getLogger("specqp.fitter")  # Creating child logger


class Peak:
    """Contains information about one peak fitted to a region.
    """
    peak_types = {
        "Gauss": ["amplitude", "center", "fwhm"],
        "Lorentz": ["amplitude", "center", "fwhm"],
        "Pseudo Voigt": ["amplitude", "center", "g_fwhm", "l_fwhm"],
        "Doniach-Sunjic": ["amplitude", "center", "g_fwhm", "l_fwhm"]
    }

    def __init__(self, x_data, y_data, popt, pcov, peak_func, peak_id, peak_type, bindingscale=True, lmfit=False):
        """Creates an instance of Peak object with X and Y data for plotting
        and fitting parameters and parameters coavriance.
        """
        self._X = x_data
        self._Y = y_data
        self._Popt = popt
        self._function = peak_func
        self._bindingscale = bindingscale
        self._id = peak_id
        self._Area = np.trapz(self._Y)
        self._PeakType = peak_type
        self._Pcov = pcov
        self._FittingErrors = []
        if not lmfit:
            for i in range(len(self._Popt)):
                try:
                    self._FittingErrors.append(np.absolute(self._Pcov[i][i])**0.5)
                except RuntimeWarning:
                    self._FittingErrors.append(0.00)
        else:
            self._FittingErrors = pcov

    def __str__(self):
        output = f"Type: {self._PeakType}"
        for i, p in enumerate(self._Popt):
            output = "\n".join((output, f"{self.peak_types[self._PeakType][i]}: {p:.4f} (+/- {self._FittingErrors[i]:.4f})"))
        output = "\n".join((output, f"Area: {self._Area:.4f}"))
        return output

    def get_covariance(self, parameter=None):
        """Returns all fitting parameters covariances or one specified by name 'peak_types'
        """
        if not parameter:
            return self._Pcov
        else:
            for i, par_name in enumerate(self.peak_types[self._PeakType]):
                if parameter == par_name:
                    return self._Pcov[i]
        fitter_logger.error(f"Couldn't get covariance from a Peak instance")

    def get_data(self):
        """Returns a list of x and y data
        """
        return [self._X, self._Y]

    def get_fitting_errors(self, parameter=None):
        """Returns fitting errors for all parameters or one specified by name 'peak_types'
        """
        if not parameter:
            return self._FittingErrors
        else:
            for i, par_name in enumerate(self.peak_types[self._PeakType]):
                if parameter == par_name:
                    return self._FittingErrors[i]
        fitter_logger.error(f"Couldn't get fitting errors from a Peak instance")

    def get_parameters(self, parameter=None):
        """Returns all fitting parameters or one specified by name from 'peak_types'
        """
        if not parameter:
            return self._Popt
        else:
            for i, par_name in enumerate(self.peak_types[self._PeakType]):
                if parameter == par_name:
                    return self._Popt[i]
        fitter_logger.error(f"Couldn't get fitting parameters from a Peak instance")

    def get_peak_area(self):
        return self._Area

    def get_peak_id(self):
        return self._id

    def get_peak_type(self):
        return self._PeakType

    def get_virtual_data(self, num=10, multiply=True):
        """Returns more (or less) data points. If multiply==False, value of 'num' is taken as desired number of ponts.
        If multiply==True, multiplies the existing number of points by value of 'num'. For plotting a smoother curve,
        for example.
        :param num: Number of desired points or multiplicator
        :param multiply: True if num is multiplicator, False if num is absolute number of points
        :return (ndarray, ndarray) x and y data with desired number of points
        """
        assert num > 0
        if multiply:
            peak_virtual_x = np.linspace(self._X[0], self._X[-1], len(self._X) * num, endpoint=True)
        else:
            peak_virtual_x = np.linspace(self._X[0], self._X[-1], num, endpoint=True)

        if self._function.__name__ == "doniach_sunjic":
            asymmetry = 'higher'
            if not self._bindingscale:
                asymmetry = 'lower'
            peak_virtual_y = self._function(peak_virtual_x, *self._Popt, asymmetry=asymmetry)
        else:
            peak_virtual_y = self._function(peak_virtual_x, *self._Popt)
        return peak_virtual_x, peak_virtual_y

    def set_peak_id(self, new_id):
        self._id = new_id


class Fitter:
    """Provides fitting possibilities for XPS regions
    """

    bg_types = (
        'constant',
        'linear',
        'square',
        'shirley'
    )

    def __init__(self, region, y_data='final', gauss_fwhm=None, bg=None):
        """Creates an object that contains information about fitting of
        a particular XPS region.
        """
        self.region = region
        self._X_data = region.get_data(column='energy')
        self._Y_data = region.get_data(column=y_data)
        # Following attributes are assigned during the fitting procedure
        self._FitLine = region.get_data(column=y_data) * 0
        self._Residuals = region.get_data(column=y_data) * 0
        self._Rsquared = 0
        self._Chisquared = 0
        self._RMS = 0
        self._Peaks = {}
        self._Bg = bg
        # The gauss widening is constant due to the equipment used in the experiment. So, if we know it,
        # we should fix this parameter in fitting.
        if gauss_fwhm:
            self._global_gauss_fwhm = gauss_fwhm * 1.0
        self._ID = region.get_id()

    def __str__(self):
        output = ""
        for i, peak in enumerate(self._Peaks.values()):
            new_line = "\n"
            if i == 0:
                new_line = ""
            output = new_line.join((output, f"--- Peak #{i+1} ---"))
            output = "\n".join((output, peak.__str__()))
        return output

    def add_peak(self, peak):
        peak_id = peak.get_peak_id()
        if peak_id not in self._Peaks:
            self._Peaks[peak_id] = peak
            return True
        else:
            fitter_logger.warning(f"Peak '{peak_id}' already exists in fitter object {self._ID}.")
            return False

    def delete_peak(self, peak_id):
        if peak_id in self._Peaks:
            del self._Peaks[peak_id]

    @staticmethod
    def get_model(model, energy, intensity, params, bindingscale=True):
        """Returns the function model according to name
        :param model: str one of specqp.Peak.peak_types
        :param energy: x axis
        :param intensity: y axis (optionally used by some models)
        :param params: parameters dictionary, e.g.
        {'amplitude': {'value': 200, 'fix': False, 'dependencetype': 'Independent',
                  'min': 10, 'max': 100000},
         'center': {'value': 22.35, 'fix': False, 'dependencetype': 'Dependent *',
                    'min': 10, 'max': 30},
         'g_fwhm': {'value': 1.15, 'fix': True, 'dependencetype': 'Independent',
                    'min': 1.1, 'max': 1.2},
         'l_fwhm': {'value': 0.5, 'fix': True, 'dependencetype': 'Independent',
                    'min': 0.4, 'max': 0.6}}
        :return: array
        """
        if model in Peak.peak_types:
            if model == "Gauss":
                return Fitter.gauss(energy, params[Peak.peak_types[model][0]]['value'],
                                    params[Peak.peak_types[model][1]]['value'],
                                    params[Peak.peak_types[model][2]]['value'])
            if model == "Lorentz":
                return Fitter.lorentz(energy, params[Peak.peak_types[model][0]]['value'],
                                      params[Peak.peak_types[model][1]]['value'],
                                      params[Peak.peak_types[model][2]]['value'])
            if model == "Pseudo Voigt":
                return Fitter.pseudo_voigt(energy, params[Peak.peak_types[model][0]]['value'],
                                           params[Peak.peak_types[model][1]]['value'],
                                           params[Peak.peak_types[model][2]]['value'],
                                           params[Peak.peak_types[model][3]]['value'])
            if model == "Doniach-Sunjic":
                asymmetry = 'higher'
                if not bindingscale:
                    asymmetry = 'lower'
                return Fitter.doniach_sunjic(energy, params[Peak.peak_types[model][0]]['value'],
                                             params[Peak.peak_types[model][1]]['value'],
                                             params[Peak.peak_types[model][2]]['value'],
                                             params[Peak.peak_types[model][3]]['value'],
                                             asymmetry=asymmetry)
        elif model in Fitter.bg_types:
            if model == 'shirley':
                return getattr(Fitter, model)(energy, intensity, params['value'])
            else:
                return getattr(Fitter, model)(energy, params['value'])
        else:
            raise KeyError(f"'{model}' is not a valid fitting model")

    @staticmethod
    def get_model_func(model):
        if model in Peak.peak_types:
            if model == "Gauss":
                return Fitter.gauss
            if model == "Lorentz":
                return Fitter.lorentz
            if model == "Pseudo Voigt":
                return Fitter.pseudo_voigt
            if model == "Doniach-Sunjic":
                return Fitter.doniach_sunjic
        elif model in Fitter.bg_types:
            if model == 'shirley':
                return Fitter.shirley
            else:
                return getattr(Fitter, model)
        else:
            raise KeyError(f"'{model}' is not a valid fitting model")

    @staticmethod
    def constant(energy, value: float, asymmetry=None):
        """Calculates constant background for simulated spectrum
        """
        return np.ones_like(energy) * value

    @staticmethod
    def linear(energy, value: float, asymmetry=None):
        """Calculates linear background for simulated spectrum
        """
        bg = np.zeros_like(energy)
        if energy[0] < energy[-1]:
            need_to_reversed = False
        else:
            need_to_reversed = True
            energy = energy[::-1]
        bg = bg + (energy - energy[0]) * value
        if need_to_reversed:
            return bg[::-1]
        else:
            return bg

    @staticmethod
    def square(energy, value: float, asymmetry=None):
        """Calculates square background for simulated spectrum
        """
        bg = np.zeros_like(energy)
        if energy[0] < energy[-1]:
            need_to_reversed = False
        else:
            need_to_reversed = True
            energy = energy[::-1]
        bg = bg + np.square((energy - energy[0])) * value
        if need_to_reversed:
            return bg[::-1]
        else:
            return bg

    @staticmethod
    def shirley(energy, intensity, value: float, asymmetry=None):
        """Calculates Shirley background for simulated spectrum
        """
        if energy[0] < energy[-1]:
            is_reversed = True
            energy = energy[::-1]
            intensity = intensity[::-1]
        else:
            is_reversed = False
        spacing = (energy[-1] - energy[0]) / (len(energy) - 1)
        output = value * spacing * (np.cumsum(intensity) - intensity.sum())
        if is_reversed:
            return output[::-1]
        return output

        # bg = np.zeros_like(energy)
        # if energy[0] < energy[-1]:
        #     need_to_reversed = False
        # else:
        #     need_to_reversed = True
        #     intensity = intensity[::-1]
        # bg = bg + (np.cumsum(intensity) - intensity) * value
        # if need_to_reversed:
        #     return bg[::-1]
        # else:
        #     return bg

    @staticmethod
    def lorentz(x, amp, cen, l_fwhm, asymmetry=None):
        gamma = l_fwhm / 2
        return amp * 1. / (np.pi * gamma * (1 + ((x - cen) / gamma) ** 2))

    @staticmethod
    def gauss(x, amp, cen, g_fwhm, asymmetry=None):
        sigma = g_fwhm / (2 * np.sqrt(2 * np.log(2)))
        return amp * 1. / np.sqrt(2 * np.pi * sigma ** 2) * np.exp(-(x - cen) ** 2 / (2 * sigma ** 2))

    @staticmethod
    def pseudo_voigt(x, amp, cen, g_fwhm, l_fwhm, asymmetry=None):
        """Returns a pseudo Voigt lineshape, used for photo-emission.
        :param x: X data
        :param amp: amplitude
        :param cen: center
        :param g_fwhm: Gauss FWHM
        :param l_fwhm: Lorentz FWHM
        :return: Pseudo Voigt line shape
        """
        f = (g_fwhm ** 5 + 2.69269 * g_fwhm ** 4 * l_fwhm + 2.42843 * g_fwhm ** 3 * l_fwhm ** 2 +
             4.47163 * g_fwhm ** 2 * l_fwhm ** 3 + 0.07842 * g_fwhm * l_fwhm ** 4 + l_fwhm ** 5) ** (1. / 5.)
        eta = 1.36603 * (l_fwhm / f) - 0.47719 * (l_fwhm / f) ** 2 + 0.11116 * (l_fwhm / f) ** 3
        pv_func = (eta * Fitter.lorentz(x, 1.0, cen, f) + (1 - eta) * Fitter.gauss(x, 1.0, cen, f))
        return amp * pv_func / np.amax(pv_func)  # Normalizing to 1

    @staticmethod
    def doniach_sunjic(x, amp, cen, g_fwhm, l_fwhm, asymmetry='higher'):
        """Returns a Doniach Sunjic asymmetric lineshape, used for photo-emission.
        Formula taken from https://lmfit.github.io/lmfit-py/builtin_models.html
        :param x: X data
        :param amp: amplitude
        :param cen: center
        :param g_fwhm: Gauss FWHM
        :param l_fwhm: Lorentz FWHM
        :param asymmetry: one of ('higher', 'lower') defines whether the secondary electrons background shall be added
        towards higher energy values (the case of binding energy representation) or lower energy values (the case
        of kinetic energy representation)
        :return: Doniach-Sunjic assimetric line shape
        """
        assert asymmetry in ('higher', 'lower')
        if x[0] > x[1]:
            reverse = True
        else:
            reverse = False
        if reverse:
            x = x[::-1]
        if asymmetry == 'higher':
            asymmetry = cen - x
        elif asymmetry == 'lower':
            asymmetry = x - cen
        else:
            fitter_logger.warning("Only 'higher' or 'lower' values are possible for asymmetry parameter.")

        sigma = g_fwhm / (2 * np.sqrt(2 * np.log(2)))
        gamma = l_fwhm / 2
        func_numerator = np.cos(np.pi * gamma / 2 + (1.0 - gamma) * np.arctan((asymmetry) / sigma))
        func_denominator = (1 + ((asymmetry) / sigma) ** 2) ** ((1.0 - gamma) / 2)
        ds_func = (amp / sigma ** (1.0 - gamma)) * func_numerator / func_denominator

        if reverse:
            return ds_func[::-1]
        return ds_func

    def _get_fitting_restrains(self, initial_params, fix_pars=None, tolerance=0.0001, boundaries=None):
        """Parses fitting restrains for multiple peaks
        :param initial_params: initial values of multiple of three (or four) parameters:
        amplitude, center, (g_fwhm - optional), and l_fwhm
        :param fix_pars: dictionary with names of parameters to fix as keys and numbers of peaks for which
        the parameters should be fixed as lists. Ex: {"cen": [1,2], "amp": [0,1,2]}
        :param tolerance: when fixing parameters some small tolerance is necessary for fitting
        :param boundaries: dictionary with names of parameters as keys and a dictionary containing lower and upper
        boundaries for the corresponding peak. Ex: {"cen": {1: [34,35], 2: [35,36]}}
        :return: Two lists: lower boundaries and upper boundaries (same order as in initial_params)
        """
        peak_pars_cnt = 0
        if len(initial_params) % 3 == 0:
            peak_pars_cnt = 3  # Fitting three-parameter function
        if len(initial_params) % 4 == 0:
            peak_pars_cnt = 4  # Fitting four-parameter function

        bounds_low = []
        bounds_high = []
        for i in range(0, len(initial_params)):
            if i % peak_pars_cnt == 0:  # Adjusting amplitude parameter boundaries
                peak_number = i // peak_pars_cnt
                if fix_pars and ("amplitude" in fix_pars):
                    if peak_number in fix_pars["amplitude"]:
                        bounds_low.append(initial_params[i] - tolerance)
                        bounds_high.append(initial_params[i] + tolerance)
                        continue
                if boundaries and ("amplitude" in boundaries):
                    if peak_number in boundaries["amplitude"]:
                        if len(boundaries["amplitude"][peak_number]) == 2 and None not in boundaries["amplitude"][peak_number]:
                            bounds_low.append(min(boundaries["amplitude"][peak_number]))
                            bounds_high.append(max(boundaries["amplitude"][peak_number]))
                            continue
                # If amplitude is not fixed and not restrained,
                # fix the lower limit at 0 and the upper limit at data_y max
                bounds_low.append(0)
                # The upper boundary of amplitude should not be lower than the value of initial guess
                if initial_params[i] < np.amax(self._Y_data):
                    bounds_high.append(np.amax(self._Y_data))
                else:
                    bounds_high.append(initial_params[i])
                continue
            if (i - 1) % peak_pars_cnt == 0:  # Adjusting center parameters boundaries
                peak_number = (i - 1) // peak_pars_cnt
                if fix_pars and ("center" in fix_pars):
                    if peak_number in fix_pars["center"]:
                        bounds_low.append(initial_params[i] - tolerance)
                        bounds_high.append(initial_params[i] + tolerance)
                        continue
                if boundaries and ("center" in boundaries):
                    if peak_number in boundaries["center"]:
                        if len(boundaries["center"][peak_number]) == 2 and None not in boundaries["center"][peak_number]:
                            bounds_low.append(min(boundaries["center"][peak_number]))
                            bounds_high.append(max(boundaries["center"][peak_number]))
                            continue
            if (i - 2) % peak_pars_cnt == 0:  # Adjusting single (or first) fwhm parameters boundaries
                peak_number = (i - 2) // peak_pars_cnt
                if fix_pars and ("g_fwhm" in fix_pars):
                    if peak_number in fix_pars["g_fwhm"]:
                        bounds_low.append(initial_params[i] - tolerance)
                        bounds_high.append(initial_params[i] + tolerance)
                        continue
                elif fix_pars and ("l_fwhm" in fix_pars):
                    if peak_number in fix_pars["l_fwhm"]:
                        bounds_low.append(initial_params[i] - tolerance)
                        bounds_high.append(initial_params[i] + tolerance)
                        continue
                if boundaries and ("g_fwhm" in boundaries):
                    if peak_number in boundaries["g_fwhm"]:
                        if len(boundaries["g_fwhm"][peak_number]) == 2 and None not in boundaries["g_fwhm"][peak_number]:
                            bounds_low.append(min(boundaries["g_fwhm"][peak_number]))
                            bounds_high.append(max(boundaries["g_fwhm"][peak_number]))
                            continue
                elif boundaries and ("l_fwhm" in boundaries):
                    if peak_number in boundaries["l_fwhm"]:
                        if len(boundaries["l_fwhm"][peak_number]) == 2 and None not in boundaries["l_fwhm"][peak_number]:
                            bounds_low.append(min(boundaries["l_fwhm"][peak_number]))
                            bounds_high.append(max(boundaries["l_fwhm"][peak_number]))
                            continue
            if peak_pars_cnt == 4:  # Adjusting second fwhm parameters boundaries
                if (i - 3) % peak_pars_cnt == 0:  # Adjusting fwhm parameters boundaries
                    peak_number = (i - 3) // peak_pars_cnt
                    if fix_pars and ("l_fwhm" in fix_pars):
                        if peak_number in fix_pars["l_fwhm"]:
                            bounds_low.append(initial_params[i] - tolerance)
                            bounds_high.append(initial_params[i] + tolerance)
                            continue
                    if boundaries and ("l_fwhm" in boundaries):
                        if peak_number in boundaries["l_fwhm"]:
                            if len(boundaries["l_fwhm"][peak_number]) == 2 and None not in boundaries["l_fwhm"][peak_number]:
                                bounds_low.append(min(boundaries["l_fwhm"][peak_number]))
                                bounds_high.append(max(boundaries["l_fwhm"][peak_number]))
                                continue
            bounds_low.append(-np.inf)
            bounds_high.append(np.inf)

        return bounds_low, bounds_high

    def fit_gaussian(self, initial_params, fix_pars=None, tolerance=0.0001, boundaries=None):
        """Fits Gaussian function to Region object based on initial values
        of three parameters (amplitude, center, and fwhm). If list with more than
        one set of three parameters is given, the function fits more than one peak.
        :param initial_params: list of initial values of parameters: amplitude, center, g_fwhm. Must contain a multiple
        of 3 values.
        :param fix_pars: dictionary with names of parameters to fix as keys and numbers of peaks for which
        the parameters should be fixed as lists. Ex: {"cen": [1,2], "amp": [0,1,2]}
        :param tolerance: when fixing parameters some small tolerance is necessary for fitting
        :param boundaries: dictionary with names of parameters as keys and a dictionary containing lower and upper
        boundaries for the corresponding peak. Ex: {"cen": {1: [34,35], 2: [35,36]}}
        """
        def _multi_gaussian(x, *args):
            cnt = 0
            func = 0
            while cnt < len(args):
                amp = args[cnt]
                cen = args[cnt + 1]
                g_fwhm = args[cnt + 2]
                func += Fitter.gauss(x, amp, cen, g_fwhm)
                cnt += 3
            return func

        if len(initial_params) % 3 != 0:
            fitter_logger.debug(f"Check the number of initial parameters in fit_gaussian method. "
                                f"Should be multiple of 3.")
            return
        bounds_low, bounds_high = self._get_fitting_restrains(initial_params, fix_pars, tolerance, boundaries)
        # Parameters and parameters covariance of the fit
        popt, pcov = curve_fit(_multi_gaussian, self._X_data, self._Y_data, p0=initial_params,
                               bounds=(bounds_low, bounds_high))

        cnt = 0
        while cnt < len(initial_params):
            peak_y = _multi_gaussian(self._X_data, popt[cnt], popt[cnt + 1], popt[cnt + 2])
            if type(peak_y) == int:
                self._Peaks[cnt // 3] = None
            else:
                self._Peaks[cnt // 3] = Peak(self._X_data, peak_y,
                                           [popt[cnt], popt[cnt+1], popt[cnt+2]],
                                           [pcov[cnt], pcov[cnt+1], pcov[cnt+2]],
                                           peak_func=_multi_gaussian,
                                           peak_id=cnt // 3, peak_type=list(Peak.peak_types.keys())[0])
            cnt += 3

        self.make_fitline()

    def fit_lorentzian(self, initial_params, fix_pars=None, tolerance=0.0001, boundaries=None):
        """Fits Lorentzian (Cauchy) function to Region object based on initial values
        of three parameters (amplitude, center, and fwhm). If list with more than
        one set of three parameters is given, the function fits more than one peak.
        :param initial_params: list of initial values of parameters: amplitude, center, l_fwhm. Must contain a multiple
        of 3 values.
        :param fix_pars: dictionary with names of parameters to fix as keys and numbers of peaks for which
        the parameters should be fixed as lists. Ex: {"cen": [1,2], "amp": [0,1,2]}
        :param tolerance: when fixing parameters some small tolerance is necessary for fitting
        :param boundaries: dictionary with names of parameters as keys and a dictionary containing lower and upper
        boundaries for the corresponding peak. Ex: {"cen": {1: [34,35], 2: [35,36]}}
        """
        def _multi_lorentzian(x, *args):
            """Creates a single or multiple Lorentzian shape taking amplitude, Center
            and FWHM parameters
            """
            cnt = 0
            func = 0
            while cnt < len(args):
                amp = args[cnt]
                cen = args[cnt + 1]
                l_fwhm = args[cnt + 2]
                func += Fitter.lorentz(x, amp, cen, l_fwhm)
                cnt += 3
            return func

        if len(initial_params) % 3 != 0:
            fitter_logger.debug(f"Check the number of initial parameters in fit_lorentzian method. "
                                f"Should be multiple of 3.")
            return
        bounds_low, bounds_high = self._get_fitting_restrains(initial_params, fix_pars, tolerance, boundaries)
        # Parameters and parameters covariance of the fit
        popt, pcov = curve_fit(_multi_lorentzian, self._X_data, self._Y_data, p0=initial_params,
                               bounds=(bounds_low, bounds_high))
        cnt = 0
        while cnt < len(initial_params):
            peak_y = _multi_lorentzian(self._X_data, popt[cnt], popt[cnt + 1], popt[cnt + 2])
            if type(peak_y) == int:
                self._Peaks[cnt // 3] = None
            else:
                self._Peaks[cnt // 3] = Peak(self._X_data, peak_y,
                                             [popt[cnt], popt[cnt+1], popt[cnt+2]],
                                             [pcov[cnt], pcov[cnt+1], pcov[cnt+2]],
                                             peak_func=_multi_lorentzian,
                                             peak_id=cnt // 3, peak_type=list(Peak.peak_types.keys())[1])
            cnt += 3
        self.make_fitline()

    def fit_pseudo_voigt(self, initial_params, fix_pars=None, tolerance=0.0001, boundaries=None):
        """Fits Pseudo-Voigt function to Region object based on initial values
        of four parameters (amplitude, center, gauss_fwhm, lorentz_fwhm). If list with more than
        one set of four parameters is given, the function fits more than one peak.
        :param initial_params: list of initial values of parameters: amplitude, center, g_fwhm, l_fwhm. Must contain a
        multiple of 4 values.
        :param fix_pars: dictionary with names of parameters to fix as keys and numbers of peaks for which
        the parameters should be fixed as lists. Ex: {"cen": [1,2], "amp": [0,1,2]}
        :param tolerance: when fixing parameters some small tolerance is necessary for fitting
        :param boundaries: dictionary with names of parameters as keys and a dictionary containing lower and upper
        boundaries for the corresponding peak. Ex: {"cen": {1: [34,35], 2: [35,36]}}
        """
        def _multi_voigt(x, *args):
            """Creates a single or multiple Voigt shape taking amplitude, Center, g_FWHM and l_FWHM parameters
            """
            cnt = 0
            func = 0
            while cnt < len(args):
                amp = args[cnt]
                cen = args[cnt + 1]
                g_fwhm = args[cnt + 2]
                l_fwhm = args[cnt + 3]
                func += Fitter.pseudo_voigt(x, amp, cen, g_fwhm, l_fwhm)
                cnt += 4
            return func

        if len(initial_params) % 4 != 0:
            fitter_logger.debug(f"Check the number of initial parameters in fit_pseudo_voigt method."
                                f"Should be multiple of 4.")
            return
        bounds_low, bounds_high = self._get_fitting_restrains(initial_params, fix_pars, tolerance, boundaries)
        # Parameters and parameters covariance of the fit
        popt, pcov = curve_fit(_multi_voigt, self._X_data, self._Y_data, p0=initial_params,
                               bounds=(bounds_low, bounds_high))

        cnt = 0
        while cnt < len(initial_params):
            peak_y = _multi_voigt(self._X_data, popt[cnt], popt[cnt + 1], popt[cnt + 2], popt[cnt + 3])
            if type(peak_y) == int:
                self._Peaks[cnt // 4] = None
            else:
                self._Peaks[cnt // 4] = Peak(self._X_data, peak_y,
                                             [popt[cnt], popt[cnt+1], popt[cnt+2], popt[cnt+3]],
                                             [pcov[cnt], pcov[cnt+1], pcov[cnt+2], pcov[cnt+3]],
                                             peak_func=_multi_voigt,
                                             peak_id=cnt // 4, peak_type=list(Peak.peak_types.keys())[2])
            cnt += 4
        self.make_fitline()

    def fit_doniach_sunjic(self, initial_params, fix_pars=None, tolerance=0.0001, boundaries=None):
        """Fits Doniach-Sunjic assimetric function (Formula taken from https://lmfit.github.io/lmfit- py/builtin_models.html)
        to Region object based on initial values of four parameters (amplitude, center, gauss_fwhm, lorentz_fwhm).
        If list with more than one set of four parameters is given, the function fits more than one peak.
        :param initial_params: list of initial values of parameters: amplitude, center, g_fwhm, l_fwhm. Must contain a
        multiple of 4 values.
        :param fix_pars: dictionary with names of parameters to fix as keys and numbers of peaks for which
        the parameters should be fixed as lists. Ex: {"cen": [1,2], "amp": [0,1,2]}
        :param tolerance: when fixing parameters some small tolerance is necessary for fitting
        :param boundaries: dictionary with names of parameters as keys and a dictionary containing lower and upper
        boundaries for the corresponding peak. Ex: {"cen": {1: [34,35], 2: [35,36]}}
        """

        def _multi_doniach_sunjic_higher(x, *args):
            """Creates a single or multiple Voigt shape taking amplitude, Center, g_FWHM and l_FWHM parameters
            """
            cnt = 0
            func = 0
            while cnt < len(args):
                amp = args[cnt]
                cen = args[cnt + 1]
                g_fwhm = args[cnt + 2]
                l_fwhm = args[cnt + 3]
                func += Fitter.doniach_sunjic(x, amp, cen, g_fwhm, l_fwhm, asymmetry='higher')
                cnt += 4
            return func

        def _multi_doniach_sunjic_lower(x, *args):
            """Creates a single or multiple Voigt shape taking amplitude, Center, g_FWHM and l_FWHM parameters
            """
            cnt = 0
            func = 0
            while cnt < len(args):
                amp = args[cnt]
                cen = args[cnt + 1]
                g_fwhm = args[cnt + 2]
                l_fwhm = args[cnt + 3]
                func += Fitter.doniach_sunjic(x, amp, cen, g_fwhm, l_fwhm, asymmetry='lower')
                cnt += 4
            return func

        fitfunc = _multi_doniach_sunjic_higher
        if not self.region.is_binding():
            fitfunc = _multi_doniach_sunjic_lower

        if len(initial_params) % 4 != 0:
            fitter_logger.debug(f"Check the number of initial parameters in fit_doniach_sunjic method."
                                f"Should be multiple of 4.")
            return
        bounds_low, bounds_high = self._get_fitting_restrains(initial_params, fix_pars, tolerance, boundaries)
        # Parameters and parameters covariance of the fit
        popt, pcov = curve_fit(fitfunc, self._X_data, self._Y_data, p0=initial_params,
                               bounds=(bounds_low, bounds_high))

        cnt = 0
        while cnt < len(initial_params):
            peak_y = fitfunc(self._X_data, popt[cnt], popt[cnt + 1], popt[cnt + 2], popt[cnt + 3])
            if type(peak_y) == int:
                self._Peaks[cnt // 4] = None
            else:
                self._Peaks[cnt // 4] = Peak(self._X_data, peak_y,
                                             [popt[cnt], popt[cnt+1], popt[cnt+2], popt[cnt+3]],
                                             [pcov[cnt], pcov[cnt+1], pcov[cnt+2], pcov[cnt+3]],
                                             peak_func=fitfunc,
                                             peak_id=cnt // 4, peak_type=list(Peak.peak_types.keys())[3])
            cnt += 4
        self.make_fitline()

    def make_fitline(self, usebg=False):
        """Calculates the total fit line including all peaks and calculates the
        residuals and r-squared.
        """
        # Calculate fit line
        for peak in self._Peaks.values():
            if peak:
                self._FitLine += peak.get_data()[1]
        if usebg and self._Bg is not None:
            for key, val in self._Bg.items():
                self._FitLine += self.get_model(key, self._X_data, self._Y_data, val)
        # Calculate residuals
        self._Residuals = self._Y_data - self._FitLine
        # Calculate R-squared
        ss_res = np.sum(self._Residuals**2)
        ss_tot = np.sum((self._Y_data - np.mean(self._Y_data))**2)
        self._Rsquared = 1 - (ss_res / ss_tot)
        # Individual standard deviation of original data
        std_d = np.sqrt((self._Y_data - np.mean(self._Y_data))**2)
        self._Chisquared = np.sum(((self._Y_data - self._FitLine)/std_d)**2)
        self._RMS = np.sum((self._Y_data - self._FitLine)**2)

    def get_bg(self):
        if self._Bg is None:
            return None
        bg_dict = {}
        for key, val in self._Bg.items():
            if type(val['value']) == float:
                bg_dict[key] = [val['value'], 0.0]
            else:
                bg_dict[key] = [val['value'].value, val['value'].stderr]
        return bg_dict

    def get_fit_line(self):
        """Returns x and y coordinates for the fit line based on all fitted peaks
        """
        if not self._FitLine.size == 0:
            return self._FitLine
        fitter_logger.warning("Can't get Fit Line from a Fitter instance. Do fit first.")

    def get_virtual_bg(self, num=10, multiply=True):
        """Returns the bg (x,y) with more (or less) data points.
        If multiply==False, value of 'num' is taken as desired number of ponts.
        If multiply==True, multiplies the existing number of points by value of 'num'. For plotting a smoother curve,
        for example.
        :param num: Number of desired points or multiplicator
        :param multiply: True if num is multiplicator, False if num is absolute number of points
        :return (ndarray, ndarray) x and y data with desired number of points
        """
        if multiply:
            bg_virtual_x = np.linspace(self._X_data[0], self._X_data[-1], len(self._X_data) * num, endpoint=True)
        else:
            bg_virtual_x = np.linspace(self._X_data[0], self._X_data[-1], num, endpoint=True)
        bg_virtual_y = np.zeros_like(bg_virtual_x)

        if self._Bg is not None:
            total_bg_y = np.zeros_like(self._Y_data)
            for key, val in self._Bg.items():
                total_bg_y += self.get_model(key, self._X_data, self._Y_data, val)

            x = list(range(len(self._X_data)))
            new_x = np.linspace(0, max(x), len(bg_virtual_x), endpoint=True)
            bg_virtual_y = np.interp(new_x, x, total_bg_y)

        return bg_virtual_x, bg_virtual_y

    def get_virtual_fitline(self, num=10, multiply=True, usebg=False):
        """Returns the fitline (x,y) with more (or less) data points.
        If multiply==False, value of 'num' is taken as desired number of ponts.
        If multiply==True, multiplies the existing number of points by value of 'num'. For plotting a smoother curve,
        for example.
        :param num: Number of desired points or multiplicator
        :param multiply: True if num is multiplicator, False if num is absolute number of points
        :return (ndarray, ndarray) x and y data with desired number of points
        """
        fitline_x, fitline_y = None, None
        for i, peak in enumerate(self._Peaks.values()):
            if peak:
                peak_data = peak.get_virtual_data(num=num, multiply=multiply)
                if i == 0:
                    fitline_x, fitline_y = peak_data
                else:
                    _, fy = peak_data
                    fitline_y += fy
        if usebg:
            _, bg_y = self.get_virtual_bg(num=num, multiply=multiply)
            fitline_y += bg_y
        return fitline_x, fitline_y

    def get_residuals(self):
        if not self._Residuals.size == 0:
            return self._Residuals
        fitter_logger.warning("Can't get residuals from a Fitter instance. Do fit first.")

    def get_rsquared(self):
        if not self._Rsquared == 0:
            return self._Rsquared
        fitter_logger.warning("Can't get R Squared from a Fitter instance. Do fit first.")

    def get_chi_squared(self):
        if not self._Chisquared == 0:
            return self._Chisquared
        fitter_logger.warning("Can't get Chi Squared from a Fitter instance. Do fit first.")

    def get_rms(self):
        if not self._RMS == 0:
            return self._RMS
        fitter_logger.warning("Can't get RMS from a Fitter instance. Do fit first.")

    # TODO add peak ID
    def get_peaks(self, peak_id=None):
        if len(self._Peaks) == 0:
            fitter_logger.warning("Can't get peaks from a Fitter instance. Do fit first.")
            return
        if peak_id is None:
            return list(self._Peaks.values())
        else:
            if peak_id in self._Peaks:
                return self._Peaks[peak_id]
            else:
                fitter_logger.warning(f"Fitter {self._ID} does not contain peak {peak_id}.")
                return

    def get_data(self):
        return [self._X_data, self._Y_data]

    def get_id(self):
        return self._ID

    def get_global_gauss_fwhm(self):
        return self._global_gauss_fwhm





import copy
import logging

import numpy as np
from lmfit import Parameters, minimize

from specqp import helpers
from specqp.fitter import Fitter, Peak

globalfitter_logger = logging.getLogger("specqp.globalfitter")  # Creating child logger


class GlobalFit:
    def __init__(self, regions, peaks_info, bg_params, y_data='final'):
        if not helpers.is_iterable(regions):
            self._Regions = [regions]
        else:
            self._Regions = regions
        self._Data = []
        self._Fitters = []
        for region in self._Regions:
            basic_data = {
                'scan': region.get_id(),
                'energy': region.get_data('energy'),
                'intensity': region.get_data(y_data)
            }
            self._Data.append(basic_data)
            self._Fitters.append(Fitter(region, y_data=y_data))
        self._FitParams = Parameters()
        self._PeaksInfo = copy.deepcopy(peaks_info)
        self._BgParams = copy.deepcopy(bg_params)
        for bgpar in self._BgParams.values():
            if bgpar['min'] is None or bgpar['min'] == "":
                bgpar['min'] = -np.inf
            if bgpar['max'] is None or bgpar['max'] == "":
                bgpar['max'] = np.inf
        self._BaseValues = {}

        self.bindingscale = True
        if not self._Regions[0].is_binding():
            self.bindingscale = False

        self.make_initial_params()

    @staticmethod
    def get_param_error(peak_name, param_data, param_name, fit_params, spectra_ind):
        if param_data['dependencetype'] == 'Dependent +':
            return (fit_params[f"{param_data['dependencebase']}_{param_name}_{spectra_ind}"].stderr ** 2 +
                    fit_params[f'{peak_name}_{param_name}_{spectra_ind}'].stderr ** 2) ** 0.5
        elif param_data['dependencetype'] == 'Dependent *':
            return ((fit_params[f"{param_data['dependencebase']}_{param_name}_{spectra_ind}"].stderr /
                     fit_params[f"{param_data['dependencebase']}_{param_name}_{spectra_ind}"].value) ** 2 +
                    (fit_params[f'{peak_name}_{param_name}_{spectra_ind}'].stderr /
                     fit_params[f'{peak_name}_{param_name}_{spectra_ind}'].value) ** 2) ** 0.5
        else:
            return fit_params[f'{peak_name}_{param_name}_{spectra_ind}'].stderr

    def get_bg_values(self, fit_params, spectra_ind):
        local_bg = copy.deepcopy(self._BgParams)
        if fit_params:
            for bg_type, params in local_bg.items():
                params['value'] = fit_params[f'bg_{bg_type}_{spectra_ind}']
        return local_bg

    def sim_spectra(self, fit_params, spectra_ind):
        """Defines the model for the fit (doniach, voigt, shirley bg, linear bg
        """
        line = np.zeros_like(self._Data[spectra_ind]['intensity'])
        bg = np.zeros_like(self._Data[spectra_ind]['intensity'])
        local_bg = self.get_bg_values(fit_params, spectra_ind)
        peaks = copy.deepcopy(self._PeaksInfo)
        for peak in peaks:
            if fit_params:
                for param_name, param_data in peak['parameters'].items():
                    param_data['value'] = fit_params[f"{peak['peakname']}_{param_name}_{spectra_ind}"]
            line += Fitter.get_model(peak['fittype'], self._Data[spectra_ind]['energy'],
                                     self._Data[spectra_ind]['intensity'], peak['parameters'],
                                     bindingscale=self.bindingscale)
        for fittype, params in local_bg.items():
            bg += Fitter.get_model(fittype, self._Data[spectra_ind]['energy'], line, params)
        line += bg
        return line, bg

    def err_func(self, fit_params):
        """ calculate total residual for fits to several data sets held
        in a 2-D array, and modeled by model function"""
        residuals = []
        for ind in range(len(self._Regions)):
            spectra, _ = self.sim_spectra(fit_params, ind)
            residuals.append(self._Data[ind]['intensity'] - spectra)
        return [item for innerlist in residuals for item in innerlist]

    def make_initial_params(self):
        """
        Creates a dictionary of parameters based on initially provided values
        :return: Dictionary of parameters for lmfit.minimize method
        """
        def get_min_first(region, pos):
            if pos == 'first':
                if region['energy'][0] > region['energy'][1]:
                    return region['intensity'][-1]
                else:
                    return region['intensity'][0]
            elif pos == 'min':
                return np.min(region['intensity'])

        # Adding background parameters to the dictionary
        try:
            for bgtype, bg_params in self._BgParams.items():
                vary = True
                if bg_params['fix']:
                    vary = False
                for ind in range(len(self._Regions)):
                    if bgtype == 'constant':
                        if bg_params['value'] in ['first', 'min']:
                            value = get_min_first(self._Data[ind], bg_params['value'])
                        else:
                            value = float(bg_params['value'])
                        if bg_params['min'] in ['first', 'min']:
                            min_ = get_min_first(self._Data[ind], bg_params['value'])
                        else:
                            min_ = float(bg_params['min'])
                        if bg_params['max'] in ['first', 'min']:
                            max_ = get_min_first(self._Data[ind], bg_params['value'])
                        else:
                            max_ = float(bg_params['max'])
                    else:
                        value = float(bg_params['value'])
                        min_ = float(bg_params['min'])
                        max_ = float(bg_params['max'])
                    self._FitParams.add(f'bg_{bgtype}_{ind}', value=value, min=min_, max=max_, vary=vary)
        except ValueError:
            self._FitParams.add(f'bg_{bgtype}_{ind}', value=0.0, min=0.0, max=0.0, vary=False)
            globalfitter_logger.warning("Check background values.")

        # Adding peak parameters to the dictionary
        common = {}
        for peak in self._PeaksInfo:
            for param_name, param_data in peak['parameters'].items():
                if param_data['min'] is None or param_data['min'] == "":
                    param_data['min'] = -np.inf
                if param_data['max'] is None or param_data['max'] == "":
                    param_data['max'] = np.inf
                vary = True
                if param_data['fix']:
                    vary = False
                for ind in range(len(self._Regions)):
                    if param_data['dependencetype'] == 'Independent':
                        self._FitParams.add(f"{peak['peakname']}_{param_name}_{ind}", value=param_data['value'],
                                            min=param_data['min'], max=param_data['max'], vary=vary)
                    elif param_data['dependencetype'] == 'Dependent +':
                        self._FitParams.add(f"{peak['peakname']}_{param_name}_{ind}_base", value=param_data['value'],
                                            min=param_data['min'],
                                            max=param_data['max'],
                                            vary=vary)
                        self._FitParams.add(f"{peak['peakname']}_{param_name}_{ind}",
                                            value=self._FitParams[f"Peak{param_data['dependencebase']}_{param_name}_{ind}"].value + param_data['value'],
                                            min=self._FitParams[f"Peak{param_data['dependencebase']}_{param_name}_{ind}"].min + param_data['min'],
                                            max=self._FitParams[f"Peak{param_data['dependencebase']}_{param_name}_{ind}"].max + param_data['max'],
                                            vary=vary,
                                            expr=f"Peak{param_data['dependencebase']}_{param_name}_{ind} + {peak['peakname']}_{param_name}_{ind}_base")
                    elif param_data['dependencetype'] == 'Dependent *':
                        self._FitParams.add(f"{peak['peakname']}_{param_name}_{ind}_base", value=param_data['value'],
                                            min=param_data['min'],
                                            max=param_data['max'],
                                            vary=vary)
                        self._FitParams.add(f"{peak['peakname']}_{param_name}_{ind}",
                                            value=self._FitParams[f"Peak{param_data['dependencebase']}_{param_name}_{ind}"].value * param_data['value'],
                                            min=self._FitParams[f"Peak{param_data['dependencebase']}_{param_name}_{ind}"].min * param_data['min'],
                                            max=self._FitParams[f"Peak{param_data['dependencebase']}_{param_name}_{ind}"].max * param_data['max'],
                                            vary=vary,
                                            expr=f"Peak{param_data['dependencebase']}_{param_name}_{ind} * {peak['peakname']}_{param_name}_{ind}_base")
                    elif param_data['dependencetype'] == 'Common':
                        if not ind == 0:
                            common[f"{peak['peakname']}_{param_name}_{ind}"] = f"{peak['peakname']}_{param_name}_{0}"
                        self._FitParams.add(f"{peak['peakname']}_{param_name}_{ind}", value=param_data['value'],
                                            min=param_data['min'], max=param_data['max'], vary=vary)

        if len(common) > 0:
            for key, val in common.items():
                self._FitParams[key].expr = val

    def fit(self):
        """Calls minimize from lmfit using the objective function and the parameters
        """
        result = minimize(self.err_func, self._FitParams, method='least_squares')
        for i, fitterobj in enumerate(self._Fitters):
            for peak in self._PeaksInfo:
                peak_pars = {}
                peak_errs = {}
                for key, val in peak['parameters'].items():
                    peak_pars[key] = result.params[f"{peak['peakname']}_{key}_{i}"].value
                    peak_errs[key] = result.params[f"{peak['peakname']}_{key}_{i}"].stderr
                asymmetry = 'higher'
                if not self.bindingscale:
                    asymmetry = 'lower'
                peak_y = Fitter.get_model_func(peak['fittype'])(fitterobj.get_data()[0], *peak_pars.values(),
                                                                asymmetry=asymmetry)
                peak = Peak(fitterobj.get_data()[0], peak_y, [*peak_pars.values()], [*peak_errs.values()],
                            peak_func=Fitter.get_model_func(peak['fittype']),
                            peak_id=peak['peakname'], peak_type=peak['fittype'],
                            bindingscale=self.bindingscale,
                            lmfit=True)
                fitterobj.add_peak(peak)
            if len(self._BgParams) > 0:
                fitterobj._Bg = self.get_bg_values(result.params, i)
                fitterobj.make_fitline(usebg=True)
            else:
                fitterobj.make_fitline(usebg=False)
        return self._Fitters



"""The module provides classes and functions responsible for loading and storing spectroscopy data.
Class Region contains the data for one region.
Class RegionCollection stores a number of Region objects
"""
import os
import ntpath
import logging
import copy
import csv
import pandas as pd
import numpy as np

from specqp import helpers

datahandler_logger = logging.getLogger("specqp.datahandler")  # Creating child logger

DATA_FILE_TYPES = (
    "scienta",
    "specs",
    "csv"
)


def load_calibration_curves(filenames, columnx='Press_03_value', columny='Press_05_value'):
    """Reads file or files using provided name(s). Checks for file existance etc.
    :param filenames: str or sequence: filepath(s)
    :param columns: str or sequence: columns to plot on y-axis
    :return:
    """
    calibration_data = {}
    if type(filenames) == str or (not type(filenames) == str and not helpers.is_iterable(filenames)):
        filenames = [filenames]
    if type(columnx) == str or (not type(columnx) == str and not helpers.is_iterable(columnx)):
        columnx = [columnx]
    if type(columny) == str or (not type(columny) == str and not helpers.is_iterable(columny)):
        columny = [columny]

    if len(columnx) != len(filenames):
        columnx = [columnx[0]] * len(filenames)
    if len(columny) != len(filenames):
        columny = [columny[0]] * len(filenames)

    for i, filename in enumerate(filenames):
        if os.path.isfile(filename):
            try:
                with open(filename, 'r') as f:
                    df = pd.read_csv(f, sep='\t')
                    calibration_data[os.path.basename(filename).rpartition(',')[2]] = \
                        (df[columnx[i]].to_numpy(), df[columny[i]].to_numpy())
            except (IOError, OSError):
                datahandler_logger.error(f"Couldn't access the file: {filename}", exc_info=True)
                continue
            except KeyError:
                datahandler_logger.error(f"The column for plotting is absent in: {filename}", exc_info=True)
                continue
    return calibration_data


def load_csv(filename, sep=None, header=None):
    """Opens and parses provided csv file returning the data and info for all regions
    as a list of Region objects. Since no experimental information is provided in the free-form text file,
    The experimental properties are set to physically miningless defaults.
    The functions assumes that the first column gives x-values and the others are y-values.
    The functions doesn't do any internal checks for file format, file access, file errors etc. and, therefore,
    should be wrapped with error handler when in use.
    """
    regions = []
    add_dimension_flag = False
    add_dimension_data = None
    df = pd.read_csv(filename, sep=sep, header=header)
    df.dropna(axis=1, inplace=True)
    for i in range(1, len(df.columns)):
        info_lines = {Region.info_entries[0]: f"Column{i}",
                              Region.info_entries[1]: "0",
                              Region.info_entries[2]: "1",
                              Region.info_entries[3]: "0",
                              Region.info_entries[4]: "Kinetic",
                              Region.info_entries[5]: str(abs(df.iat[0, 0] - df.iat[1, 0])),
                              Region.info_entries[6]: "1",
                              Region.info_entries[7]: filename.rpartition('/')[2],
                              Region.info_entries[8]: ""}

        region_conditions = {"Comments": ""}
        # Create a Region object for the current region
        region_id = f"{info_lines[Region.info_entries[7]]} : {info_lines[Region.info_entries[0]]}"
        regions.append(Region(df.iloc[:, 0].to_numpy(), df.iloc[:, i].to_numpy(), id_=region_id,
                              add_dimension_flag=add_dimension_flag, add_dimension_data=add_dimension_data,
                              info=info_lines, conditions=region_conditions))

    return regions


def load_scienta_txt(filename, regions_number_line=1, first_region_number=1):
    """Opens and parses provided scienta.txt file returning the data and info for all regions
    as a list of Region objects. Variable 'regions_number_line' gives the
    number of the line in the scienta file where the number of regions is given
    (the line numbering starts with 0 and by default it is the line number 1 that
    contains the information).
    The functions doesn't do any internal checks for file format, file access, file errors etc. and, therefore,
    should be wrapped with error handler when in use.
    """

    def parse_scienta_file_info(lines_):
        """Helper function that parses the list of lines acquired from Scienta.txt
        file info block and returns 'info' dictionary {property: value}
        """
        info = {}
        for line_ in lines_:
            if '=' in line_:
                line_content = line_.strip().split('=', 1)
                # Modify the file name
                if line_content[0].strip() == "File":
                    line_content[1] = ntpath.basename(line_content[1]).split('.', 1)[0]
                info[line_content[0].strip()] = line_content[1].strip()
        return info

    with open(filename) as f:
        lines = f.read().splitlines()

    # Dictionary that contains the map of the file, where the name of the section (specific region) is
    # the key and the lists of first and last indices of Region, Info and Data subsections are the values
    # Example: {"Region 1": [[0, 2], [3, 78], [81, 180]]}
    #                        Region    Info      Data
    file_map = {}
    # Reading the number of regions from the specified line of the file
    regions_number = int(lines[regions_number_line].split("=", 1)[1])
    # We make a list of region objects even for just one region
    regions = []
    # Temporary counter to know the currently treated region number in the for-loop below
    cnt = first_region_number  # (=1 by default)
    # Temporary list variables to store the first and the last indices of the
    # region, the info and the data blocks for every region
    region_indices = []
    info_indices = []
    data_indices = []

    # Parsing algorithm below assumes that the file structure is constant and the blocks follow the sequence:
    # [Region N] - may contain info about add-dimension mode
    # [Info N] - important info
    # [Data N] - data
    for i, line in enumerate(lines):
        if f"[Region {cnt}]" in line:
            region_indices.append(i + 1)
            # If it is not the first region, than the data section of the previous
            # region ends on the previous line
            if cnt > 1:
                data_indices.append(i - 1)
            continue
        if f"[Info {cnt}]" in line:
            info_indices.append(i + 1)
            region_indices.append(i - 1)  # Region section ends on the previous line
            continue
        if f"[Data {cnt}]" in line:
            data_indices.append(i + 1)
            info_indices.append(i - 1)  # Info section ends on the previous line
            # If it is the last region, the current Data section is the last. The last line of the file is the last
            # line of the current Data section. Else, we go to the next region.
            if cnt == regions_number:
                data_indices.append(len(lines) - 1)
                break
            else:
                cnt += 1

    # Resetting region number counter to 1 to start again from the first region and do the mapping procedure
    cnt = first_region_number  # (=1 by default)
    # Iterate through every pair [begin, end] of Region, Info and Data indices to populate the mapping dictionary
    for j in range(1, len(region_indices), 2):
        file_map[f"Region {cnt}"] = [[region_indices[j-1], region_indices[j]],
                                     [info_indices[j-1], info_indices[j]],
                                     [data_indices[j-1], data_indices[j]]]
        cnt += 1

    # Iterating through the mapping dictionary
    for val in file_map.values():
        # Variables which are necessary for the case of add-dimension region
        add_dimension_flag = False
        add_dimension_data = None

        # Region block of the current region
        region_block = lines[val[0][0]:val[0][1] + 1]  # List of lines within [begin, end] indices including end-values
        for line in region_block:
            # If the region is measured in add-dimension mode
            if "Dimension 2 size" in line:
                add_dimension_flag = True
                add_dimension_data = []
                break

        # Info block of the current region
        info_lines = parse_scienta_file_info(lines[val[1][0]:val[1][1]+1])
        # Not all info entries are important for data analysis,
        # Choose only important ones
        info_lines_revised = {Region.info_entries[0]: info_lines["Region Name"],
                              Region.info_entries[1]: info_lines["Pass Energy"],
                              Region.info_entries[2]: info_lines["Number of Sweeps"],
                              Region.info_entries[3]: info_lines["Excitation Energy"],
                              Region.info_entries[4]: info_lines["Energy Scale"],
                              Region.info_entries[5]: info_lines["Energy Step"],
                              Region.info_entries[6]: info_lines["Step Time"],
                              Region.info_entries[7]: info_lines["File"],
                              Region.info_entries[8]: f"{info_lines['Date']} {info_lines['Time']}"}

        region_conditions = {"Comments": info_lines["Comments"]}

        # Data block of the current region
        data_block = lines[val[2][0]:val[2][1]+1]
        energy, counts = [], []
        for i, line in enumerate(data_block):
            if not line.strip():
                continue  # Skip empty lines
            else:
                xy = list(map(float, line.split()))
                energy.append(xy[0])
                if not add_dimension_flag:
                    counts.append(xy[1])
                # If add-dimension mode is one, there will be a number of columns instead of just two
                # We read them row by row and then transpose the whole thing to get columns
                else:
                    row_counts_values = []
                    # We skip the first value every time because it contains energy which is the same for all columns
                    for ncol in range(1, len(xy)):
                        row_counts_values.append(xy[ncol])
                    counts.append(sum(row_counts_values))  # 'counts' list value contains integrated rows
                    add_dimension_data.append(row_counts_values)

        if add_dimension_data:
            add_dimension_data = list(map(list, zip(*add_dimension_data)))  # Transpose
        # Create a Region object for the current region
        region_id = f"{info_lines_revised[Region.info_entries[7]]} : {info_lines_revised[Region.info_entries[0]]}"
        regions.append(Region(energy, counts, id_=region_id,
                              add_dimension_flag=add_dimension_flag, add_dimension_data=add_dimension_data,
                              info=info_lines_revised, conditions=region_conditions))

    return regions


# TODO: Add possibility to read add_dimension files
def load_specs_xy(filename):
    """Opens and parses provided SPECS file returning the data and info for recorded
    region as a list of Region objects in order to be consistent with the Scienta
    loading routine.
    """
    def parse_spec_file_info(lines_):
        """Parses the list of lines read from SPEC.xy file info block
        and returns 'info' dictionary
        """
        info = {}
        for line_ in lines_:
            if ':' in line_:
                line_ = line_.strip(' #')
                line_content = line_.split(':', 1)
                info[line_content[0].strip()] = line_content[1].strip()

        return info

    with open(filename) as f:
        lines = f.read().splitlines()

    # Basic parsing based on the appearance of SPECS files
    energy, counts = [], []
    info_lines = []
    for line in lines:
        line = line.strip()
        if not line:
            continue  # Scip empty lines
        elif line.startswith('#'):
            if line.strip(' #'):  # Scip empty info lines
                info_lines.append(line)  # Save info lines
        else:
            xy = line.split()
            x = float(xy[0].strip())
            y = float(xy[1].strip())
            if y > 0:
                energy.append(x)
                counts.append(y)

    regions = []
    # The file might be empty, then we ignore it
    if energy:
        # Switch from list to dictionary
        info_lines = parse_spec_file_info(info_lines)

        # Check which energy scale is used:
        if info_lines["Energy Axis"] == "Kinetic Energy":
            info_lines["Energy Axis"] = "Kinetic"  # To make it consistent with Scienta routine
        else:
            info_lines["Energy Axis"] = "Binding"

        # Not all info entries are important for data analysis,
        # Choose only important ones
        info_lines_revised = {Region.info_entries[0]: info_lines["Region"],
                              Region.info_entries[1]: info_lines["Pass Energy"],
                              Region.info_entries[2]: info_lines["Number of Scans"],
                              Region.info_entries[3]: info_lines["Excitation Energy"],
                              Region.info_entries[4]: info_lines["Energy Axis"],
                              Region.info_entries[5]: str(abs(energy[-1] - energy[0]) / int(info_lines["Values/Curve"])),
                              Region.info_entries[6]: info_lines["Dwell Time"],
                              Region.info_entries[7]: filename.rpartition('.')[0].rpartition('/')[2],
                              Region.info_entries[8]: info_lines["Acquisition Date"]}

        region_conditions = {"Comments": info_lines["Comment"]}
        # Create a Region object for the current region
        regions.append(Region(energy, counts, info=info_lines_revised, conditions=region_conditions))
    return regions


class Region:
    """Class Region contains the data and info for one measured region, e.g. C1s
    """
    info_entries = (
        "Region Name",          # 0
        "Pass Energy",          # 1
        "Sweeps Number",        # 2
        "Excitation Energy",    # 3
        "Energy Scale",         # 4
        "Energy Step",          # 5
        "Dwell Time",           # 6
        "File Name",            # 7
        "Date"                  # 8
    )

    region_flags = (
        "energy_shift_corrected",  # 0
        "binding_energy_flag",     # 1
        "fermi_flag",              # 2
        "sweeps_normalized",       # 3
        "add_dimension_flag"       # 4
    )

    def __init__(self, energy, counts,
                 add_dimension_flag=False, add_dimension_data=None,
                 info=None, conditions=None, excitation_energy=None,
                 id_=None, fermi_flag=False, flags=None):
        """
        :param energy: goes for energy (X) axis
        :param counts: goes for counts (Y) axis
        :param id_: ID of the region (usually the string "filename : regionname")
        :param fermi_flag: True if the region stores the measurement of the Fermi level
        :param add_dimension_flag: True if the region stores two-dimensional data (in this case, the values in the
        'counts' and 'final' columns are obtained by integration over the second dimension. At the same time
        separate columns 'counts0', 'counts1'... and 'final0', 'final1'... are added
        to the region object and contain information for time-distributed scans)
        :param add_dimension_data: list of lists containing separate data for time-distributed scans
        :param info: Info about the region is stored as a dictionary {property: value} based on info_entries tuple
        :param conditions: Experimental conditions are stored as a dictionary {property: value}
        :param excitation_energy: Photon energy used in the experiment
        :param flags: dictionary of flags if already processed data is to be imported to a new Region object
        """
        # The main attribute of the class is pandas dataframe
        self._data = pd.DataFrame(data={'energy': energy, 'counts': counts}, dtype=float)
        self._applied_corrections = []
        self._info = info
        self._id = id_
        # Experimental conditions
        self._conditions = conditions
        # Excitation energy
        if excitation_energy:
            self._excitation_energy = float(excitation_energy)
            self._info[Region.info_entries[3]] = str(float(excitation_energy))
        else:
            self._excitation_energy = None
        if flags and (len(flags) == len(Region.region_flags)):
            self._flags = flags
        else:
            if flags:
                datahandler_logger.warning("Incorrect dictionary of flags when creating a region object")
            # Default values for flags
            self._flags = {
                    Region.region_flags[0]: False,
                    Region.region_flags[1]: None,
                    Region.region_flags[2]: fermi_flag,
                    Region.region_flags[3]: False,
                    Region.region_flags[4]: add_dimension_flag
                    }
        # Check which energy scale is used:
        if self._info:  # Info can be None
            if self._info[Region.info_entries[4]] == "Binding":
                self._flags[Region.region_flags[1]] = True
            else:
                self._flags[Region.region_flags[1]] = False
            # If info is available for the region and the ID is not assigned,
            # take string "FileName : RegionName" as ID
            if not self._id:
                self.set_id(f"{self._info[Region.info_entries[7]]} : {self._info[Region.info_entries[0]]}")
        if Region.region_flags[4] and add_dimension_data:
            self._add_dimension_scans_number = len(add_dimension_data)
            for i, data_set in enumerate(add_dimension_data):
                self.add_column(f'counts{i}', data_set)
                self.add_column(f'final{i}', data_set)
        else:
            self._add_dimension_scans_number = 1

        # 'final' column is the main y-data column for plotting. At the beginning it is identical to the 'counts' values
        self.add_column('final', self._data["counts"])

        # A backup, which can be used to restore the initial state of the region object.
        # If the region is a dummy region that doesn't contain any data, the .copy() action is not available
        try:
            self._data_backup = self._data.copy()
            self._info_backup = self._info.copy()
            self._flags_backup = self._flags.copy()
        except AttributeError:
            datahandler_logger.info(f"A dummy region has been created", exc_info=True)

    def __str__(self):
        """Prints the info read from the data file. Possible to add keys of the Info dictionary to be printed
        """
        return self.get_info_string()

    # TODO: write region subtraction routine
    def __sub__(self, other_region):
        """Returns the difference of the "final" column of two instances
        """
        return self._data["final"] - other_region.get_data("final")

    def add_column(self, column_label, array, overwrite=False):
        """Adds one column to the data object assigning it the name 'column_label'.
        Choose descriptive labels. If label already exists but 'overwrite' flag
        is set to True, the method overwrites the data in the column.
        """
        if column_label in self._data.columns and not overwrite:
            datahandler_logger.warning(f"Column '{column_label}' already exists in {self.get_id()}"
                                       "Pass overwrite=True to overwrite the existing values.")
            return
        self._data[column_label] = array

    def add_correction(self, correction: str):
        self._applied_corrections.append(correction)

    @staticmethod
    def addup_regions(regions, truncate=True):
        """
        Takes two or more non-adddimension regions and combines them in a single region adding up sweeps and results.
        The method checks for regions' energy spans and truncates to the overlapping interval if truncate=True. Skips
        non-matching regions if truncate=False
        NOTE: The method doesn't check for regions compatibility, i.e. conditions, photon energy, etc.
        NOTE: Neither does the method accounts for previously made corrections. Therefore, should be used with
        freshly loaded regions in order to avoid mistakes.
        :param regions: iterable through a number of regions
        :param truncate=True: if regions of different length are combined and truncate == True, the method
        adds them up only in the overlapping region and returns truncated regions as the result
        :return: Region object
        """

        def _find_overlap(a, b):
            ab_intersect = np.in1d(a, b).nonzero()[0]
            ba_intersect = np.in1d(b, a).nonzero()[0]
            if len(ab_intersect) == 0 or len(ba_intersect) == 0:
                return None, None
            return [ab_intersect[0], ab_intersect[-1]], [ba_intersect[0], ba_intersect[-1]]

        if not helpers.is_iterable(regions):
            datahandler_logger.warning(f"Regions should be in an iterable object to be concatenated")
            return None
        region = copy.deepcopy(regions[0])
        for i in range(1, len(regions)):
            energy1 = region.get_data('energy')
            energy2 = regions[i].get_data('energy')
            # Check that the regions can be concatenated, i.e. if they have full/partial overlap
            indxs1, indxs2 = _find_overlap(energy1, energy2)
            if (not indxs1) or (not indxs2):
                continue
            if indxs1 == indxs2 and len(energy1) == len(energy2):  # If regions fully overlap
                region._data['counts'] += regions[i].get_data('counts')
                region._data['final'] += regions[i].get_data('final')
            elif indxs1 != indxs2 and truncate:
                if not (indxs1[0] == 0 and indxs1[1] == len(energy1) + 1):
                    region.crop_region(start=energy1[indxs1[0]], stop=energy1[indxs1[1]], changesource=True)
                region._data['counts'] += regions[i].get_data('counts')[indxs2[0]:indxs2[1]]
                region._data['final'] += regions[i].get_data('final')[indxs2[0]:indxs2[1]]
            else:
                datahandler_logger.warning(f"Regions {region.getid()} and {regions[i].get_id()} have different length")
                continue
            region._info[Region.info_entries[2]] = str(int(region.get_info(Region.info_entries[2])) +
                                                       int(regions[i].get_info(Region.info_entries[2])))
        return region

    @staticmethod
    def bin_add_dimension(region, nbins, drop_remainder=False):
        """
        Takes an add-dimension region and returns a list of non-add-dimension regions that are binned
        subdimensions.
        :param region: add-dimension region
        :param nbins: number of bins
        :param drop_remainder: if the last bin is not full, drop it
        :return: add-dimension region with binned subdimensions of the initial region
        """
        if not region.is_add_dimension():
            return region
        if region.is_add_dimension() and region.get_add_dimension_counter() < nbins:
            return region
        scans_per_bin = region.get_add_dimension_counter() // nbins
        bincnt = 0
        binned_add_dimension_data = []
        binned_add_dimension_columns = {}
        for column in region.get_data_columns():
            if 'energy' not in column and 'counts' not in column and 'final' not in column:
                col_base_name = ''.join(filter(lambda x: x.isalpha(), column))
                if col_base_name not in binned_add_dimension_columns:
                    binned_add_dimension_columns[col_base_name] = region.get_data(col_base_name)

        column_names = list(binned_add_dimension_columns.keys())
        for i in range(region.get_add_dimension_counter()):
            if i == nbins * scans_per_bin and drop_remainder:
                break
            if i == bincnt * scans_per_bin:
                bin_intensity = region.get_data(f'counts{i}')
                for column_name in column_names:
                    if f'{column_name}{i}' in region.get_data_columns():
                        binned_add_dimension_columns[f'{column_name}{i}'] = region.get_data(f'{column_name}{i}')
            else:
                bin_intensity += region.get_data(f'counts{i}')
                for column_name in column_names:
                    if f'{column_name}{i}' in region.get_data_columns():
                        binned_add_dimension_columns[f'{column_name}{i}'] += region.get_data(f'{column_name}{i}')
            if i == (bincnt + 1) * scans_per_bin - 1 or i == region.get_add_dimension_counter() - 1:
                if i == region.get_add_dimension_counter() - 1 and region.get_add_dimension_counter() % nbins != 0:
                    # If we don't wont to drop non-full lust bin, we need to scale it to the same number of
                    # sweeps as for other bins
                    bin_intensity *= scans_per_bin / (region.get_add_dimension_counter() % nbins)
                binned_add_dimension_data.append(bin_intensity)
                bin_intensity = None
                bincnt += 1
        binned_region_info = copy.deepcopy(region.get_info())
        binned_region_info['Sweeps Number'] = int(binned_region_info['Sweeps Number']) * scans_per_bin
        binned_region = Region(region.get_data('energy'), region.get_data('counts'),
                               add_dimension_flag=True, add_dimension_data=binned_add_dimension_data,
                               info=binned_region_info, conditions=region.get_conditions(),
                               excitation_energy=region.get_excitation_energy(),
                               id_=f"{region.get_id()}",
                               fermi_flag=region.get_flags()[Region.region_flags[2]],
                               flags=copy.deepcopy(region.get_flags()))
        binned_region._add_dimension_scans_number = len(binned_add_dimension_data)
        binned_region._applied_corrections = region.get_corrections()
        binned_region._flags_backup = region._flags_backup
        binned_region._flags[Region.region_flags[4]] = True  # Add-dimension flag
        for key, val in binned_add_dimension_columns.items():
            binned_region.add_column(key, val)
        binned_region.add_column('final', region.get_data('final'), overwrite=True)
        return binned_region

    def correct_energy_shift(self, shift):
        if not self._flags[Region.region_flags[0]]:  # If not already corrected
            self._data['energy'] += shift
            self._flags[Region.region_flags[0]] = True
            self._applied_corrections.append("Energy shift corrected")
        else:
            datahandler_logger.info(f"The region {self._id} has already been energy corrected.")

    def crop_region(self, start=None, stop=None, changesource=False):
        """Returns a copy of the region with the data within [start, stop] interval
        on 'energy' x-axis. Interval is given in real units of the data. If start or
        stop or both are not specified the method takes first (or/and last) values.
        If changesource flag is True, the original region is cropped, if False -
        the copy of original region is cropped and returned.
        """
        if start is None and stop is None:
            return

        x_values = self._data['energy']
        if start is not None and stop is not None:
            if (x_values.iloc[0] > x_values.iloc[-1] and start < stop) or (x_values.iloc[-0] < x_values.iloc[-1] and start > stop):
                start, stop = stop, start

        first_index = 0
        last_index = self._data.index.values[-1]
        if start is not None:
            for i in x_values.index:
                if i > 0:
                    if (x_values[i - 1] <= start <= x_values[i]) or (x_values[i - 1] >= start >= x_values[i]):
                        first_index = i
        if stop is not None:
            for i in x_values.index:
                if i > 0:
                    if (x_values[i - 1] <= stop <= x_values[i]) or (x_values[i - 1] >= stop >= x_values[i]):
                        last_index = i

        if changesource:
            self._data = self._data.truncate(before=first_index, after=last_index)
            # Reset indexing after truncation so that it starts again with 0
            self._data.reset_index(drop=True, inplace=True)
            return

        tmp_region = copy.deepcopy(self)
        tmp_region._data = tmp_region._data.truncate(before=first_index, after=last_index)
        tmp_region._data.reset_index(drop=True, inplace=True)
        return tmp_region

    def get_add_dimension_counter(self):
        """
        :return: Number of separate scans in add-dimension region
        """
        return self._add_dimension_scans_number

    def get_conditions(self, entry=None, as_string=False):
        """Returns experimental conditions as a dictionary {"entry": value} or
        the value of the specified entry.
        """
        if self._conditions:
            if entry:
                return self._conditions[entry]
            elif as_string:
                cond_string = ""
                for val in self._conditions.values():
                    if not bool(cond_string):  # If string is empty
                        cond_string = "".join([cond_string,val])
                    else:
                        cond_string = "; ".join([cond_string, val])
                return cond_string
            return self._conditions
        else:
            datahandler_logger.warning(f"The region {self._id} doesn't contain information about conditions.")

    def get_corrections(self, as_string=False):
        """Returns either the list or the string (if as_string=True) of corrections that has been applied to the region
        """
        if not as_string:
            return self._applied_corrections
        else:
            if not self._applied_corrections:
                return "Not corrected"
            else:
                output = ""
                for cor in self._applied_corrections:
                    if not output:
                        output = "".join([output, cor])
                    else:
                        output = "; ".join([output, cor])
                return output

    def get_data(self, column=None):
        """Returns pandas DataFrame with data columns. If column name is
        provided, returns 1D numpy.ndarray of specified column.
        """
        if column:
            return self._data[column].to_numpy()
        return self._data

    def get_data_columns(self, add_dimension=True) -> list:
        # If we want to get all add_dimension columns or the region is not add_dimension, return all columns
        if add_dimension or not self._flags[self.region_flags[4]]:
            return self._data.columns.to_list()
        else:
            # Return only main columns, the ones that don't contain digits
            cols = [col for col in self._data.columns if not any([char.isdigit() for char in col])]
            return cols

    def get_excitation_energy(self):
        return self._excitation_energy

    def get_flags(self):
        """Returns the dictionary of flags
        """
        return self._flags

    def get_id(self):
        return self._id

    def get_info(self, parameter=None):
        """Returns 'info' dictionary {"name": value} or the value of specified
        parameter.
        """
        if parameter:
            return self._info[parameter]
        return self._info

    def get_info_string(self, include_conditions=True, *args):
        """Returns info string with the information about the region
        Possible to add keys of the Info dictionary to be printed
        """
        output = ""
        if not self._info:
            output = f"No info available for region {self.get_id()}"
        else:
            # If no specific arguments provided, add everything to the output
            if len(args) == 0:
                for key, val in self._info.items():
                    output = "\n".join((output, f"{key}: {val}"))
                if include_conditions:
                    for key, val in self._conditions.items():
                        output = "\n".join((output, f"{key}: {val}"))
            else:
                # Add only specified parameters
                for arg in args:
                    if arg in self._info:
                        output = "\n".join((output, f"{arg}: {self._info[arg]}"))
                    elif include_conditions and arg in self._conditions:
                        output = "\n".join((output, f"{arg}: {self._conditions[arg]}"))
                    else:
                        datahandler_logger.warning(f"Parameter {arg} is not known for region {self._id}")
        return output

    def invert_energy_scale(self):
        """Changes the energy scale of the region from the currently defined to
        the alternative one. From kinetic to binding energy or from binding to kinetic energy.
        """
        self._data['energy'] = -1 * self._data['energy'] + self._excitation_energy
        self._flags[Region.region_flags[1]] = not self._flags[Region.region_flags[1]]

        # We need to change "Energy Scale" info entry also
        if self._flags[Region.region_flags[1]]:
            self._info[Region.info_entries[4]] = "Binding"
        else:
            self._info[Region.info_entries[4]] = "Kinetic"

    def invert_to_binding(self):
        """Changes the energy scale of the region from kinetic to binding energy.
        """
        if not self._flags[Region.region_flags[1]]:
            self.invert_energy_scale()

    def invert_to_kinetic(self):
        """Changes the energy scale of the region from binding to kinetic energy.
        """
        if self._flags[Region.region_flags[1]]:
            self.invert_energy_scale()

    def is_add_dimension(self):
        return self._flags[self.region_flags[4]]

    def is_binding(self):
        return self._flags[self.region_flags[1]]

    def is_dummy(self):
        if len(self._data['energy']) == 0:
            return True
        return False

    def is_energy_corrected(self):
        return self._flags[self.region_flags[0]]

    def is_sweeps_normalized(self):
        return self._flags[self.region_flags[3]]

    def make_final_column(self, parent_column, overwrite=False):
        """Populates the 'final' column with the values from the column "parent_column", which contains processed data.
        Populates all 'finalN' columns with values from 'parent_columnN' for add_dimension regions.
        """
        self.add_column('final', self._data[parent_column], overwrite)
        if self.is_add_dimension():
            for i in range(self._add_dimension_scans_number):
                if f'{parent_column}{i}' in list(self._data):
                    self.add_column(f'final{i}', self._data[f'{parent_column}{i}'], overwrite)

    def normalize_by_sweeps(self, column='final'):
        """
        Normalizes 'column' column by sweeps and stores the result in the new column 'sweepsNormalized'.
        For add_dimension regions normalizes all columns 'columnN' and creates new columns 'sweepsNormalizedN'
        NOTE: The routine also takes into account the dwell time per point
        :return:
        """
        # If not yet normalized by sweeps
        if not self._flags[self.region_flags[3]]:
            if self._info and (Region.info_entries[2] in self._info):
                if not self._flags[self.region_flags[4]]:
                    self._data['sweepsNormalized'] = self._data[column] / (float(self._info[Region.info_entries[2]]) *
                                                                           float(self._info[Region.info_entries[6]]))
                else:
                    # This many sweeps in each add_dimension measurement
                    sweeps_per_set = self._info[Region.info_entries[2]]
                    for i in range(self._add_dimension_scans_number):
                        if f'{column}{i}' in self._data.columns:
                            self._data[f'sweepsNormalized{i}'] = (self._data[f'{column}{i}'] /
                                                                  (float(sweeps_per_set) *
                                                                   float(self._info[Region.info_entries[6]])))
                    self._data['sweepsNormalized'] = (self._data[column] /
                                                      (float(int(sweeps_per_set) * self._add_dimension_scans_number) *
                                                       float(self._info[Region.info_entries[6]])))
                self._flags[self.region_flags[3]] = True
                return True
        return False

    @staticmethod
    def read_csv(filename):
        """Reads csv file and returns Region object. Values of flags and info
        is retrieved from the comment lines marked with '#' simbol at the beginning
        of the file.
        """
        try:
            with open(filename, 'r') as region_file:
                data_start = False
                flags = {}
                info = {}
                conditions = {}
                _id = None
                _scans_cnt = 1
                applied_c = []
                while not data_start:
                    line = region_file.readline()
                    if line.strip() == "[DATA]":
                        data_start = True
                        continue
                    elif '#ID' in line:
                        _id = line.replace("#ID", "").strip()
                    elif '#AD' in line:
                        _scans_cnt = int(line.replace("#AD", "").strip())
                    elif '#F' in line:
                        key, _, val = line.replace("#F", "").strip().partition("=")
                        flags[key] = val
                    elif '#I' in line:
                        key, _, val = line.replace("#I", "").strip().partition("=")
                        info[key] = val
                    elif '#C' in line:
                        key, _, val = line.replace("#C", "").strip().partition("=")
                        conditions[key] = val
                    elif "#AC":
                        applied_c = line.replace("#AC", "").strip().split(';')
                        applied_c = [ac.strip() for ac in applied_c]
                data = pd.read_csv(region_file, sep='\t')

                region = Region([], [], info=info, excitation_energy=float(info[Region.info_entries[3]]), conditions=conditions, id_=_id, flags=flags)
                region._data = data
                region._add_dimension_scans_number = _scans_cnt
                region._applied_corrections = applied_c
                region._data_backup = data.copy()
                region._info_backup = info.copy()
                region._flags_backup = flags.copy()
                return region
        except (IOError, OSError):
            datahandler_logger.warning(f"Can't access the file {filename}", exc_info=True)
            return False

    def reset_region(self):
        """Removes all the changes made to the Region and restores the initial
        "counts" and "energy" columns together with the _info, _flags
        """
        if self._data_backup and self._info_backup and self._flags_backup:
            self._data = self._data_backup.copy()
            self._info = self._info_backup.copy()
            self._flags = self._flags_backup.copy()
            self._applied_corrections = []
        else:
            datahandler_logger.warning("Attempt to reset a dummy region. Option is not available for dummy regions.")

    def save_xy(self, file, cols='final', add_dimension=True, headers=True):
        """Saves Region object as csv file with 'energy' and other specified columns. If add_dimension region
        is provided and 'add_dimension' variable is True, saves 'energy' column and specified columns for all sweeps.
        :param cols: Sequence, String, 'all'. Which columns of the Region dataframe to save along with the energy column
        :param add_dimension: Saves all relevant columns for all sweeps of add_dimension region
        :param file: File handle
        :param headers: Include the columns headers in the file
        :return: True if successful, False otherwise
        """
        if cols == 'all':
            cols = self.get_data_columns(add_dimension)
        else:
            if not (type(cols) != str and helpers.is_iterable(cols)):
                cols = [cols]
            else:
                cols = list(cols)  # Convert possible sequences to list
            # Check if columns exist in self._data and report the missing columns
            missing_cols = cols.copy()
            cols = [c for c in cols if c in self._data.columns]
            missing_cols = list(set(missing_cols) - set(cols))
            for mc in missing_cols:
                datahandler_logger.warning(f"Can't save column '{mc}' in region {self._id}.")
            # If the region is add-dimension and we want to save it as add_dimension -> save all relevant columns
            if self._flags[self.region_flags[4]] and add_dimension:
                add_dimension_cols = []
                for col in cols:
                    add_dimension_cols += [c for c in self._data.columns if col in c]
                cols = add_dimension_cols
            if 'energy' not in cols:
                cols = ['energy'] + cols

        try:
            self._data[cols].round(2).to_csv(file, header=headers, index=False, sep='\t')
        except (OSError, IOError):
            datahandler_logger.error(f"Can't write the file {file.name}", exc_info=True)
            return False
        print(f"Created: {file.name}")
        return True

    def save_as_file(self, file, cols='all', add_dimension=True, details=True, headers=True):
        """Saves Region object as csv file with all columns. If details==True, saves also info, conditions and
        :param details: Bool
        :param file: File handle
        :param cols: Which columns to write to file
        :param add_dimension: If True, saves data for all sweeps, if False, saves only main columns
        :param headers: Include the columns headers in the file
        :return: True if successful, False otherwise
        """
        if not details:
            if not self.save_xy(file, cols=cols, add_dimension=add_dimension, headers=headers):
                return False
        else:
            try:
                file.write(f"#ID {self._id}\n")
                if self.is_add_dimension():
                    file.write(f"#AD {self._add_dimension_scans_number}\n")
                for key, value in self._flags.items():
                    file.write(f"#F {key}={value}\n")
                for key, value in self._info.items():
                    file.write(f"#I {key}={value}\n")
                for key, value in self._conditions.items():
                    file.write(f"#C {key}={value}\n")
                file.write(f"#AC {self.get_corrections(as_string=True)}\n")
                file.write("[DATA]\n")
            except (OSError, IOError):
                datahandler_logger.error(f"Can't write the file {file.name}", exc_info=True, sep='\t')
                return False
            if not self.save_xy(file, cols=cols, add_dimension=add_dimension, headers=headers):
                return False
        return True

    @staticmethod
    def separate_add_dimension(region):
        """
        Takes an add-dimension region and returns a list of non-add-dimension regions that are single subdimensions.
        :param region: add-dimension region
        :return: list of non-add-dimension regions that are subdimensions of the initial region
        """
        separated_regions = []
        if not region.is_add_dimension():
            return region
        for i in range(region.get_add_dimension_counter()):
            dimension = Region(region.get_data('energy'), region.get_data(f'counts{i}'),
                               add_dimension_flag=False, add_dimension_data=None,
                               info=region.get_info(), conditions=region.get_conditions(),
                               excitation_energy=region.get_excitation_energy(),
                               id_=f"{region.get_id()} : Sweep {i}",
                               fermi_flag=region.get_flags()[Region.region_flags[2]],
                               flags=copy.deepcopy(region.get_flags()))
            dimension._add_dimension_scans_number = 1
            dimension._applied_corrections = region.get_corrections()
            dimension._flags_backup = region._flags_backup
            dimension._flags[Region.region_flags[4]] = False  # Not add-dimension any longer
            for column in region.get_data_columns():
                if str(i) in column and 'energy' not in column and 'counts' not in column and 'final' not in column:
                    col_base_name = ''.join(filter(lambda x: x.isalpha(), column))
                    if f"{col_base_name}{i}" == column:
                        dimension.add_column(col_base_name, region.get_data(column))
            dimension.add_column('final', region.get_data(f'final{i}'), overwrite=True)
            separated_regions.append(dimension)
        return separated_regions

    def set_conditions(self, conditions, overwrite=False):
        """Set experimental conditions as a dictionary {"Property": Value}. If conditions with the same names
        already exist will skip/overwrite depending on the overwrite value.
        """
        # If some conditions already exist
        if self._conditions:
            for key, val in conditions.items():
                if key in self._conditions and not overwrite:
                    continue
                self._conditions[key] = val
        else:
            self._conditions = conditions

    def set_excitation_energy(self, excitation_energy):
        """Set regions's excitation energy.
        """
        self._excitation_energy = float(excitation_energy)
        self._info[Region.info_entries[3]] = str(float(excitation_energy))

    def set_fermi_flag(self):
        self._flags[Region.region_flags[2]] = True

    def set_id(self, region_id):
        self._id = region_id


class RegionsCollection:
    """Keeps track of the list of regions being in work simultaneously in the GUI or the batch mode
    """
    def __init__(self, regions=None):
        """
        :param regions: List of region objects (can be also single object in the list form, e.g. [obj,])
        """
        self.regions = {}
        if regions:
            for region in regions:
                self.regions[region.get_id()] = region

    def add_regions(self, new_regions):
        """Adds region objects. Checks for duplicates and rejects adding if already exists.
        :param new_regions: List of region objects (can be also single object in the list form, e.g. [obj,])
        :return: list of IDs for regions that were added
        """
        ids = []
        duplicate_ids = []  # For information purposes
        for new_region in new_regions:
            new_id = new_region.get_id()
            if self.is_duplicate(new_id):
                duplicate_ids.append(new_id)
                continue
            else:
                ids.append(new_id)
                self.regions[new_id] = new_region
        if duplicate_ids:
            datahandler_logger.warning(f"Regions are already loaded: {duplicate_ids}")
        if ids:
            return ids

    def add_regions_from_file(self, file_path, file_type=DATA_FILE_TYPES[0]):
        """Adds region objects after extracting them from the file.
        Checks for duplicates and rejects adding if already exists.
        :param file_path: Absolute path to the data file from which the regions shall be extracted
        :param file_type: File type to be processed
        :return: list of IDs for regions loaded from the file
        """
        ids = []
        try:
            if file_type == DATA_FILE_TYPES[0]:
                ids = self.add_regions(load_scienta_txt(file_path))
            elif file_type == DATA_FILE_TYPES[1]:
                ids = self.add_regions(load_specs_xy(file_path))
            elif file_type == DATA_FILE_TYPES[2]:
                ids = self.add_regions(load_csv(file_path))
        except OSError:
            datahandler_logger.error(f"Couldn't access the file {file_path}", exc_info=True)
            return None
        except UnicodeDecodeError:
            datahandler_logger.error(f"Couldn't decode the file {file_path}", exc_info=True)
            return None
        except ValueError:
            datahandler_logger.error(f"The file {file_path} has unexpected characters", exc_info=True)
            return None
        except Exception:
            datahandler_logger.error(f"The file {file_path} is corrupted", exc_info=True)
            return None
        return ids

    def get_by_id(self, region_id):
        if not type(region_id) == str and helpers.is_iterable(region_id):  # Return multiple regions
            return [self.regions[reg_id] for reg_id in region_id if reg_id in self.regions]
        else:
            if region_id in self.regions:
                return self.regions[region_id]

    def get_ids(self):
        """Returns the list of IDs for regions in RegionCollection object
        :return: list of IDs
        """
        return list(self.regions.keys())

    def get_regions(self):
        return self.regions.values()

    def is_duplicate(self, id_):
        if id_ in self.regions:
            return True
        else:
            return False



"""Provides functions for handling and fitting the data
"""
import os
import logging
import scipy as sp
import numpy as np
from scipy.optimize import curve_fit


helpers_logger = logging.getLogger("specqp.helpers")  # Creating child logger


def is_iterable(obj):
    try:
        _ = (e for e in obj)
        return True
    except TypeError:
        return False


def fit_fermi_edge(region, initial_params, column="final", add_column=True, overwrite=True):
    """Fits error function to fermi level scan. If add_column flag
    is True, adds the fitting results as a column to the Region object.
    NOTE: Overwrites the 'fitFermi' column if already present.
    Returns a list [shift, fittingError]
    """

    # f(x) = s/(exp(-1*(x-m)/(8.617*(10^-5)*t)) + 1) + a*x + b
    def error_func(x, a0, a1, a2, a3):
        """Defines a complementary error function of the form
        (a0/2)*sp.special.erfc((a1-x)/a2) + a3
        """
        return (a0 / 2) * sp.special.erfc((a1 - x) / a2) + a3

    if not region.get_flags()["fermi_flag"]:
        helpers_logger.warning(f"Can't fit the error func to non-Fermi region {region.get_id()}")
        return

    # Parameters and parameters covariance of the fit
    popt, pcov = curve_fit(error_func,
                           region.get_data(column='energy'),
                           region.get_data(column=column),
                           p0=initial_params)

    if add_column:
        region.add_column("fitFermi", error_func(region.get_data(column='energy'),
                                                popt[0],
                                                popt[1],
                                                popt[2],
                                                popt[3]),
                          overwrite=overwrite)
    # Return parameters and their uncertainties
    return [popt, np.sqrt(np.diag(pcov))]


def subtract_linear_bg(region, y_data='final', manual_bg=None, by_min=False, add_column=True, overwrite=True):
    """Calculates the linear background using left and right ends of the region
    or using the minimum on Y-axis and the end that is furthest from the minimum
    on the X-axis. Manual background can be provided by passing approximate intervals
    like [[730, 740], [760,770]]; In that case the average values within these
    intervals will be calculated and assigned to the end points of the line
    describing the background.
    """

    def calculate_line(min_position):
        """Helper function to calculate the line cross the whole
        spectrum given coordinates of two points and "left" or "right" value
        of min_position to know in which direction the line must be extended
        """
        # School algebra to calculate the line coordinates given two points
        if min_position == "right":  # Line from left end to min
            x1 = energy[0]
            x2 = energy[counts_min_index]
            y1 = counts[0]
            y2 = counts_min
            slope = (y1 - y2) / (x1 - x2)
            b = (x1 * y1 - x2 * y1) / (x1 - x2)
            line_end = slope * energy[-1] + b
            return np.linspace(counts[0], line_end, len(energy))
        elif min_position == "left":
            x1 = energy[counts_min_index]
            x2 = energy[-1]
            y1 = counts_min
            y2 = counts[-1]
            slope = (y1 - y2) / (x1 - x2)
            b = (x1 * y1 - x2 * y1) / (x1 - x2)
            line_end = slope * energy[0] + b
            return np.linspace(line_end, counts[-1], len(energy))

    def calculate_manual_bg(x, y, x_intervals):
        left_interval = x_intervals[0]
        right_interval = x_intervals[1]
        # Checking that left is left and right is right. Swop otherwise.
        if x[0] > x[-1]:
            if left_interval[0] < right_interval[0]:
                left_interval = x_intervals[1]
                right_interval = x_intervals[0]

        left_bg_values = []
        right_bg_values = []

        first_left_index = 0
        last_left_index = len(x) - 1
        first_right_index = 0
        last_right_index = len(x) - 1
        for i in range(1, len(x)):
            if (x[i - 1] >= left_interval[0] >= x[i]) or (x[i - 1] <= left_interval[0] <= x[i]):
                first_left_index = i
            if (x[i - 1] >= left_interval[1] >= x[i]) or (x[i - 1] <= left_interval[1] <= x[i]):
                last_left_index = i
            if (x[i - 1] >= right_interval[0] >= x[i]) or (x[i - 1] <= right_interval[0] <= x[i]):
                first_right_index = i
            if (x[i - 1] >= right_interval[1] >= x[i]) or (x[i - 1] <= right_interval[1] <= x[i]):
                last_right_index = i

        left_background = y[first_left_index:last_left_index + 1]
        left_average = np.mean(left_background)
        # sum(left_background)/float(len(left_background))
        right_background = y[first_right_index:last_right_index + 1]
        right_average = np.mean(right_background)
        # sum(right_background)/float(len(right_background))

        return [left_average, right_average]

    counts = region.get_data(column=y_data).tolist()
    energy = region.get_data(column="energy").tolist()

    if by_min:
        counts_min = min(counts)
        counts_min_index = counts.index(counts_min)
        # If minimum lies closer to the right side of the region
        if counts_min_index > len(energy) // 2:
            background = calculate_line("right")
        else:
            background = calculate_line("left")
    else:
        if manual_bg:
            line_end_values = calculate_manual_bg(energy, counts, manual_bg)
            background = np.linspace(line_end_values[0], line_end_values[1], len(energy))
        else:
            background = np.linspace(counts[0], counts[-1], len(energy))

    if add_column:
        region.add_column("no_lin_bg", counts - background, overwrite=overwrite)

    return background


def subtract_shirley(region, y_data='final', tolerance=1e-5, maxiter=50, add_column=True, overwrite=True):
    """Calculates shirley background. Adopted from https://github.com/schachmett/xpl
    Author Simon Fischer <sfischer@ifp.uni-bremen.de>"
    """
    def get_shirley_bg(energy, counts, tolerance=1e-5, maxiter=50):
        if energy[0] < energy[-1]:
            is_reversed = True
            energy = energy[::-1]
            counts = counts[::-1]
        else:
            is_reversed = False
        background = np.ones(energy.shape) * counts[-1]
        integral = np.zeros(energy.shape)
        spacing = (energy[-1] - energy[0]) / (len(energy) - 1)
        subtracted = counts - background
        ysum = subtracted.sum() - np.cumsum(subtracted)
        for i in range(len(energy)):
            integral[i] = spacing * (ysum[i] - 0.5 * (subtracted[i] + subtracted[-1]))
        iteration = 0
        while iteration < maxiter:
            subtracted = counts - background
            integral = spacing * (subtracted.sum() - np.cumsum(subtracted))
            bnew = ((counts[0] - counts[-1]) * integral / integral[0] + counts[-1])
            if np.linalg.norm((bnew - background) / counts[0]) < tolerance:
                background = bnew.copy()
                break
            else:
                background = bnew.copy()
            iteration += 1
        if iteration >= maxiter:
            return None
        output = background
        if is_reversed:
            output = background[::-1]
        return output

    energy = region.get_data(column="energy")
    counts = region.get_data(column=y_data)
    bg = get_shirley_bg(energy, counts, tolerance, maxiter)
    if bg is None:
        helpers_logger.warning(f"{region.get_id()} - Shirley background calculation failed due to excessive iterations")
        bg = counts * 0
    if add_column:
        corrected = counts - bg
        if np.amin(corrected) < 0:
            corrected += np.absolute(np.amin(corrected))
        region.add_column("no_shirley", corrected, overwrite=overwrite)
    if region.is_add_dimension():
        main_output = bg
        add_dimension_output = []
        for i in range(region.get_add_dimension_counter()):
            if f'{y_data}{i}' in region.get_data().columns:
                counts = region.get_data(column=f'{y_data}{i}')
                bg = get_shirley_bg(energy, counts, tolerance, maxiter)
                if bg is None:
                    helpers_logger.warning(f"{region.get_id()} : Add-dimension line {i} - "
                                           f"Shirley background calculation failed due to excessive iterations")
                    bg = counts * 0
                add_dimension_output.append(bg)
                if add_column:
                    corrected = counts - bg
                    if np.amin(corrected) < 0:
                        corrected += np.absolute(np.amin(corrected))
                    region.add_column(f"no_shirley{i}", corrected, overwrite=True)
        return main_output, add_dimension_output
    return bg


def calculate_linear_and_shirley(region, y_data='counts', shirleyfirst=True, by_min=False, tolerance=1e-5, maxiter=50,
                                 add_column=True, overwrite=True):
    """If shirleyfirst=False, calculates the linear background using left and
    right ends of the region or using the minimum and the end that is furthest
    from the minimum if by_min=True. Then calculates shirley background.
    If shirleyfirst=True, does shirley first and linear second.
    """
    if y_data in list(region.get_data()):
        counts = region.get_data(column=y_data)
    else:
        counts = region.get_data(column='counts')

    if shirleyfirst:
        shirley_bg = subtract_shirley(region, tolerance=tolerance, maxiter=maxiter, add_column=add_column)
        linear_bg = subtract_linear_bg(region, y_data='no_shirley', by_min=by_min, add_column=add_column)
    else:
        linear_bg = subtract_linear_bg(region, by_min=by_min, add_column=add_column)
        shirley_bg = subtract_shirley(region, y_data="no_lin_bg", tolerance=tolerance, maxiter=maxiter,
                                      add_column=add_column)
    background = linear_bg + shirley_bg
    if add_column:
        region.add_column("no_lin_no_shirley", counts - background, overwrite=overwrite)

    return background


def smoothen(region, y_data='counts', interval=3, add_column=True):
    """Smoothes intensity averaging the data within the given interval"""
    intensity = region.get_data(column=y_data)
    odd = int(interval / 2) * 2 + 1
    even = int(interval / 2) * 2
    cumsum = np.cumsum(np.insert(intensity, 0, 0))
    avged = (cumsum[odd:] - cumsum[:-odd]) / odd
    for _ in range(int(even / 2)):
        avged = np.insert(avged, 0, avged[0])
        avged = np.insert(avged, -1, avged[-1])

    if add_column:
        region.add_column("averaged", avged, overwrite=True)

    return avged


def normalize(region, y_data='final', const=None, add_column=True):
    """Normalize counts by maximum. If const is given, normalizes by this number. If add_dimension region is received
    normalizes the main (integrated) columns 'counts', 'final' etc. as usual. Other columns are normalized as well in
    the case const is None (takes max of every add_dimesion column) or normalize by a constant if a list of
    corresponding constants is provided. In case single constant is provided, add_dimension columns 'final0', 'final1'
    etc. are not normalized.
    """
    # If we want to use other column than "final" for calculations
    if not region.is_add_dimension():
        if is_iterable(const):
            return False
        else:
            y = region.get_data(column=y_data)
            if const:
                output = y / float(const)
            else:
                output = y / float(max(y))
            if add_column:
                region.add_column("normalized", output, overwrite=True)
                return True
            else:
                return output
    else:
        if not const:
            for i in range(region.get_add_dimension_counter()):
                if f'{y_data}{i}' in list(region.get_data()):
                    y = region.get_data(column=f'{y_data}{i}')
                    region.add_column(f"normalized{i}", y / float(np.amax(y)), overwrite=True)
                    y = region.get_data(column=f'{y_data}')
                    region.add_column(f"normalized", y / float(np.amax(y)), overwrite=True)
        # If const is provided it should be iterable containing values for every corresponding add-dimension column
        # elif isinstance(const, Iterable):
        else:
            output = []
            if is_iterable(const):
                if len(const) == region.get_add_dimension_counter():
                    main_res = region.get_data(column=y_data) / float(np.mean(const))
                    for i, constant in enumerate(const):
                        res = region.get_data(column=f'{y_data}{i}') / float(constant)
                        if add_column:
                            region.add_column("normalized", main_res, overwrite=True)
                            region.add_column(f"normalized{i}", res, overwrite=True)
                        else:
                            output.append(res)
                    if add_column:
                        return True
                    else:
                        return main_res, output
                else:
                    helpers_logger.warning(f"Add-dimension data in region {region.get_id()} was not normalized.")
                    return False
            else:
                main_res = region.get_data(column=y_data) / float(const)
                for i in range(region.get_add_dimension_counter()):
                    res = region.get_data(column=f'{y_data}{i}') / float(const)
                    if add_column:
                        region.add_column("normalized", main_res, overwrite=True)
                        region.add_column(f"normalized{i}", res, overwrite=True)
                    else:
                        output.append(res)
                if add_column:
                    return True
                else:
                    return main_res, output


def normalize_by_background(region, start, stop, y_data='counts', add_column=True):
    """Correct counts by the average background level given by the interval [start, stop]
    """
    # If we want to use other column than "counts" for calculations
    counts = region.get_data(column=y_data)
    energy = region.get_data(column='energy')

    first_index = 0
    last_index = len(counts) - 1

    for i in range(0, len(energy)):
        if i > 0:
            if (energy[i - 1] <= start <= energy[i]) or (energy[i - 1] >= start >= energy[i]):
                first_index = i
            if (energy[i - 1] <= stop <= energy[i]) or (energy[i - 1] >= stop >= energy[i]):
                last_index = i

    output = counts / float(np.mean(counts[first_index:last_index]))

    if add_column:
        region.add_column("bgnormalized", output, overwrite=True)
    return output


def normalize_group(regionscollection, y_data: str = 'final',
                    const: float = None, add_column: bool = True) -> bool:
    """Normalize y-axis of all regions in the RegionsCollection by the maximum y-value of all included regions.
       If const is given, normalizes by this number. If add_dimension region is received
       normalizes the main (integrated) columns 'counts', 'final' etc. only.
    """
    # If we want to use other column than "counts" for calculations
    regions = regionscollection.get_regions()
    if const:
        if is_iterable(const) and len(const) == len(regions):
            for i, region in enumerate(regions):
                output = normalize(region, y_data=y_data, const=const[i], add_column=False)
                if region.is_add_dimension():
                    output = output[0]
                if add_column:
                    region.add_column("groupnormalized", output, overwrite=True)
            return True
        elif is_iterable(const) and len(const) != len(regions):
            helpers_logger.warning(f"Regions collection was not normalized because number of normalization"
                                   f"constants was not equal to number of regions.")
            return False
        elif not is_iterable(const):
            for region in regions:
                output = normalize(region, y_data=y_data, const=const, add_column=False)
                if region.is_add_dimension():
                    output = output[0]
                if add_column:
                    region.add_column("groupnormalized", output, overwrite=True)
            return True

    # Normalization coefficient for all regions (max y-value of all regions from y_data column)
    allmax = max([np.max(region.get_data(column=y_data)) for region in regions if len(region.get_data(column=y_data)) > 0])
    for region in regions:
        output = normalize(region, y_data=y_data, const=allmax, add_column=False)
        if region.is_add_dimension():
            output = output[0]
        if add_column:
            region.add_column("groupnormalized", output, overwrite=True)
    return True


def shift_by_background(region, interval, y_data='final', add_column=True):
    """Correct counts by the average background level given by the interval [start, stop]
    """
    # If we want to use other column than "counts" for calculations
    counts = region.get_data(column=y_data)
    energy = region.get_data(column="energy")

    first_index = 0
    last_index = len(counts) - 1

    for i in range(0, len(energy)):
        if i > 0:
            if (energy[i - 1] <= interval[0] <= energy[i]) or (energy[i - 1] >= interval[0] >= energy[i]):
                first_index = i
            if (energy[i - 1] <= interval[1] <= energy[i]) or (energy[i - 1] >= interval[1] >= energy[i]):
                last_index = i

    main_output = counts - float(np.mean(counts[first_index:last_index]))
    add_dimension_outputs = []
    if add_column:
        region.add_column("bgshifted", main_output, overwrite=True)
    if region.is_add_dimension():
        for i in range(region.get_add_dimension_counter()):
            if f'{y_data}{i}' in region.get_data().columns:
                counts = region.get_data(column=f'{y_data}{i}')
                output = counts - float(np.mean(counts[first_index:last_index]))
                add_dimension_outputs.append(output)
                if add_column:
                    region.add_column(f"bgshifted{i}", output, overwrite=True)
    if add_dimension_outputs:
        return main_output, add_dimension_outputs
    else:
        return main_output


def ask_path(folder_flag=True, multiple_files_flag=False):
    """Makes a tkinter dialog for choosing the folder if folder_flag=True
    or file(s) otherwise. For multiple files the multiple_files_flag should
    be True.
    """
    # This method is almost never used, so the required imports are locally called
    import tkinter as tk
    from tkinter import filedialog

    root = tk.Tk()
    root.withdraw()
    path = os.getcwd()
    if folder_flag:  # Open folder
        path = filedialog.askdirectory(parent=root, initialdir=path, title='Please select directory')
    else:  # Open file
        if multiple_files_flag:
            path = filedialog.askopenfilenames(parent=root, initialdir=path, title='Please select data files')
            path = root.tk.splitlist(path)
        else:
            path = filedialog.askopenfilename(parent=root, initialdir=path, title='Please select data file')
    root.destroy()
    return path



import logging
from matplotlib.figure import Figure

from specqp import helpers
from specqp.datahandler import Region
# import helpers

plotter_logger = logging.getLogger("specqp.plotter")  # Creating child logger


def _get_arrays(region, x_data='energy', y_data='final'):
    """
    :param region: Region object
    :param x_data: region's column name
    :param y_data: region's column name
    :return: ndarrays of specified columns' values
    """
    if x_data in list(region.get_data()):
        x = region.get_data(column=x_data)
    else:
        x = region.get_data(column='energy')
    if y_data in list(region.get_data()):
        y = region.get_data(column=y_data)
    else:
        y = region.get_data(column='final')

    return x, y


def _make_label(region, legend_features=None):
    def _add_to_label(lbl: str, new_lbl_entry: str):
        if not bool(lbl):  # If label is empty
            lbl = "".join([lbl, new_lbl_entry])
        else:
            lbl = " : ".join([lbl, new_lbl_entry])
        return lbl

    label = ""
    if 'ID' in legend_features:
        label = _add_to_label(label, f"{region.get_id()}")
    for entry in legend_features:
        if entry in Region.info_entries:
            label = _add_to_label(label, f"{region.get_info(entry)}")
    features = region.get_conditions()
    if "Conditions" in legend_features and features:
        label = _add_to_label(label, f"{region.get_conditions(as_string=True)}")
    elif legend_features and features:
        for legend_feature in legend_features:
            if legend_feature in features:
                label = _add_to_label(label, features[legend_feature])
    elif legend_features and not features:
        plotter_logger.info(f"Conditions for region {region.get_id()} are not known.")
    return label


def plot_add_dimension(region, axs, x_data='energy', y_data='final', invert_x=True, log_scale=False, y_offset=0.0,
                       global_y_offset=0.0, scatter=False, label=None, color=None, title=False, font_size=12,
                       legend=True, legend_features=None, legend_pos='best'):
    """

    :param region: region object
    :param axs: Matplotlib axis object
    :param x_data: type string. Name of the region's data column to plot on X-axis
    :param y_data: type string. Name of the region's data column to plot on Y-axis
    :param invert_x: True/False for the X-axis (Binding/Kinetic energy representation)
    :param log_scale: True/False for the Y-axis
    :param y_offset: offset for each next curve along the Y-axis
    :param global_y_offset: common offset for all curves along the Y-axis
    :param scatter: scatter or solid line-shape
    :param label: label of curves. Can be list of labels for all add-dimension curves.
    :param color: color of curves. Can be list of colors for all add-dimension curves.
    :param title: General title of the plot
    :param font_size:
    :param legend: True/Flase
    :param legend_features: What info to add to the labels from the region's infp section.
    For example, legend_features=('Temperature',)
    :param legend_pos: Legend position according to matplotlib rules. For example, 'best'
    :return: None
    """
    if not region.is_add_dimension():
        # plot_region(region, axs, x_data=x_data, y_data=y_data, invert_x=invert_x, log_scale=log_scale, y_offset=y_offset,
        #             scatter=scatter, label=label, color=color, title=title, font_size=font_size,
        #             legend=legend, legend_features=legend_features, legend_pos=legend_pos)
        plotter_logger.warning(f"Region {region.get_id()} was not plotted because it is not add-dimension.")
        return

    n_curves = region.get_add_dimension_counter()
    if label and not helpers.is_iterable(label):
        label = [f'{label} : sweep {i}' for i in range(1, n_curves + 1)]
    elif (label and len(label) != n_curves) or not label:
        #plotter_logger.info(f"Labels for region {region.get_id()} plotting were set to defaults.")
        label = _make_label(region, legend_features=legend_features)
        label = [f'{label} : sweep {i}' for i in range(1, n_curves + 1)]

    if color and not helpers.is_iterable(color):
        color = [color for _ in range(1, n_curves + 1)]
    elif (color and len(color) != n_curves) or not color:
        color = [None for _ in range(1, n_curves + 1)]

    for i in range(n_curves):
        x, y = _get_arrays(region, x_data=f'{x_data}{i}', y_data=f'{y_data}{i}')
        if global_y_offset:
            y += global_y_offset
        _plot_curve(x, y, region, axs, invert_x=invert_x, log_scale=log_scale, y_offset=i*y_offset,
                    scatter=scatter, label=label[i], color=color[i], title=title, font_size=font_size,
                    legend=legend, legend_features=legend_features, legend_pos=legend_pos)


def plot_region(region, axs, x_data='energy', y_data='final', invert_x=True, log_scale=False, y_offset=0,
                scatter=False, label=None, color=None, title=True, font_size=12,
                legend=True, legend_features=None, legend_pos='best'):
    """Plotting spectrum with matplotlib using given axes and a number of optional arguments. legend_features parameter
    allows for adding distinguishing features to each plotted curve taking their values from Region._conditions.
    Example: legend_features=('Temperature',)
    """
    x, y = _get_arrays(region, x_data, y_data)
    _plot_curve(x, y, region, axs, invert_x=invert_x, log_scale=log_scale, y_offset=y_offset,
                scatter=scatter, label=label, color=color, title=title, font_size=font_size,
                legend=legend, legend_features=legend_features, legend_pos=legend_pos)


def _plot_curve(x, y, region, axs, invert_x=True, log_scale=False, y_offset=0.0,
                scatter=False, label=None, color=None, title=True, font_size=12,
                legend=True, legend_features=None, legend_pos='best'):

    if label is None:
        label = _make_label(region, legend_features=legend_features)
    # If we want scatter plot
    if not bool(label):
        label = " "
    if scatter:
        axs.scatter(x, y + y_offset, s=7, c=color, label=label)
    else:
        axs.plot(x, y + y_offset, color=color, label=label)
    axs.tick_params(axis='both', which='both', labelsize=font_size)
    if legend:
        if legend_pos == 'lower center':
            axs.set_ylim(ymin=-1 * max(y.tolist()))
            axs.tick_params(
                axis='y',  # changes apply to the y-axis
                which='both',  # both major and minor ticks are affected
                left=False,  # ticks along the left edge are off
                right=False)  # ticks along the right edge are off
        axs.legend(fancybox=True, framealpha=0, loc=legend_pos, prop={'size': font_size})
    if title:
        if region.is_sweeps_normalized():
            title_str = f"Pass: {region.get_info('Pass Energy')} | File: {region.get_info('File Name')}"
        else:
            if region.is_add_dimension():
                title_str = f"Pass: {region.get_info('Pass Energy')} " \
                    f" | Sweeps: {int(region.get_info('Sweeps Number'))*region.get_add_dimension_counter()}" \
                    f" | File: {region.get_info('File Name')}"
            else:
                title_str = f"Pass: {region.get_info('Pass Energy')} | Sweeps: {region.get_info('Sweeps Number')}" \
                    f" | File: {region.get_info('File Name')}"
        axs.set_title(title_str, fontsize=font_size)

    #   Stiling axes
    x_label_prefix = "Binding"
    if region.get_info("Energy Scale") == "Kinetic":
        x_label_prefix = "Kinetic"

    axs.set_xlabel(f"{x_label_prefix} energy (eV)", fontsize=font_size)
    axs.set_ylabel("Counts (a.u.)", fontsize=font_size)

    # Inverting x-axis if desired and not yet inverted
    if invert_x and not axs.xaxis_inverted():
        axs.invert_xaxis()

    if log_scale:
        axs.set_yscale('log')

    # Switch to scientific notation when y-axis numbers get bigger than 4 digits
    if max(y) > 9999:
        axs.ticklabel_format(axis='y', style='sci', scilimits=(0, 0))


def plot_peak(peak, axs, y_offset=0.0, label=None, color=None, fill=True, legend=True, legend_pos='best', font_size=12):
    """
    Plotting one peak from Fitter object
    :param peak: Peak object
    :param axs: matplotlib.figure.axes object
    :param y_offset: vertical offset of the plot
    :param label: legend entry
    :param color: color
    :param fill: fills the peak area with half-transparent color (same as the color of the curve)
    :param legend: Enable/disable legend
    :param legend_pos: legend position
    :param font_size: font_size for all text within the axes object
    :return: None
    """
    peak_line = peak.get_data()
    if not label:
        label = f"Cen: {peak.get_parameters('center'):.2f}; LorentzFWHM: {peak.get_parameters('l_fwhm'):.2f}"

    plot_peak_xy(peak_line[0], peak_line[1], axs, y_offset=y_offset, label=label, color=color,
                 fill=fill, legend=legend, legend_pos=legend_pos, font_size=font_size)


def plot_peak_xy(x, y, axs, y_offset=0.0, label=None, baseline=None,
                 color=None, fill=True, legend=True, legend_pos='best', font_size=12):
    """Plotting one peak by given (x, y) points
    :param x: x coordinates
    :param y: y coordinates
    :param axs: matplotlib.figure.axes object
    :param y_offset: vertical offset of the plot
    :param label: legend entry
    :param color: color
    :param fill: fills the peak area with half-transparent color (same as the color of the curve)
    :param legend: Enable/disable legend
    :param legend_pos: legend position
    :param font_size: font_size for all text within the axes object
    :return: None
    """
    axs.plot(x, y + y_offset, color=color, label=label)
    if fill:
        # Fill the peak shape with color that is retrieved from the last plotted line
        if baseline is not None:
            axs.fill_between(x, baseline + y_offset, y + y_offset, facecolor=axs.get_lines()[-1].get_color(), alpha=0.3)
        else:
            axs.fill_between(x, y.min() + y_offset, y + y_offset, facecolor=axs.get_lines()[-1].get_color(), alpha=0.3)
    if legend:
        axs.legend(fancybox=True, framealpha=0, loc=legend_pos, prop={'size': font_size})


def plot_fit(fitter, axs, y_offset=0, label=None, color='black', legend=True, legend_pos='best',
             addresiduals=True, font_size=12):
    """Plotting fit line with pyplot using given plt.figure and a number of optional arguments
    """
    if not label:
        label = f"fit {fitter.get_id()}"
    axs.plot(fitter.get_data()[0], fitter.get_fit_line() + y_offset, linestyle='--', color=color, label=label)
    # Add residuals if specified
    if addresiduals:
        axs.plot(fitter.get_data()[0], fitter.get_residuals() + y_offset, linestyle=':', alpha=1, color='black',
                label=f"Chi^2 = {fitter.get_chi_squared():.2f}\nRMS = {fitter.get_rms():.2f}")
    if legend:
        axs.legend(fancybox=True, framealpha=0, loc=legend_pos, prop={'size': font_size})


def stylize_axes(ax):
    ax.set_aspect('auto')
    ax.set_facecolor('None')
    ax.grid(which='both', axis='both', color='grey', linestyle=':')
    ax.spines['bottom'].set_color('black')
    ax.spines['left'].set_color('black')
    ax.tick_params(axis='x', colors='black')
    ax.tick_params(axis='y', colors='black')
    ax.yaxis.label.set_color('black')
    ax.xaxis.label.set_color('black')


class SpecqpPlot(Figure):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import sys
import os
import logging
from logging.handlers import RotatingFileHandler

# import service
# from gui import main as call_gui
from specqp import service
from specqp.gui import main as call_gui


specqp_logger = logging.getLogger("specqp")


def initialize_logging():
    """Setting up the main logger for the app
    """
    specqp_logger.setLevel(logging.ERROR)  # Main level filter for log messages (with DEBUG all messages are evaluated)

    # Setting up console output with more information to be logged
    console_handler = logging.StreamHandler()
    console_handler.setLevel(logging.ERROR)  # Secondary level filter for log messages in console

    # Create the log directory if doesn't exist
    directory = os.path.dirname(service.service_vars["LOG_FILE_NAME"])
    if not os.path.exists(directory):
        os.makedirs(directory)
    # Setting up logfile output with only errors and critical errors to be logged
    # Set rotating handlers so that the logfile doesn't grow forever
    file_handler = RotatingFileHandler(service.service_vars["LOG_FILE_NAME"], maxBytes=1000000, backupCount=3)
    file_handler.setLevel(logging.ERROR)  # Secondary level filter for log messages in file

    formatter = logging.Formatter("%(asctime)s (%(levelname)s in %(name)s): %(message)s", "%Y-%m-%d %H:%M:%S")
    console_handler.setFormatter(formatter)
    file_handler.setFormatter(formatter)

    specqp_logger.addHandler(console_handler)
    specqp_logger.addHandler(file_handler)


def parse_batch_file(filename, sections=None):
    guidelines = []
    fl = filename
    if not os.path.isfile(filename):
        filename = os.getcwd() + '/' + filename
        if not os.path.isfile(filename):
            specqp_logger.error(f"Can't find the source text file {fl}. The file was not loaded.")
            return [None]
    with open(filename, 'r') as f:
        lines = f.readlines()
        if sections is None:
            guidelines = [line for line in lines if line[0] == 'F']
            return guidelines
        else:
            for section in sections:
                section_start = [i + 1 for i, line in enumerate(lines) if f"[{section}]" in line]
                section_end = [i for i, line in enumerate(lines) if f"[/{section}]" in line]
                if len(section_start) != len(section_end):
                    specqp_logger.error(f"Every section of the instruction file should begin with"
                                        "'[sectionname]' label and end with [/sectionname] label. "
                                        "The file {fl} was not loaded.")
                    return [None]
                for i in range(len(section_start)):
                    guidelines += [line for line in lines[section_start[i]:section_end[i]] if line[0] == 'F']
            return guidelines


# TODO: write the logics for the batch mode
def main(*args, **kwargs):
    """Defines the behavior of the app if run with flags and/or other parameters
    """
    # If no command line arguments or only '-gui' flag provided, run blank GUI
    if (len(args) == 0 and len(kwargs) == 0) or \
       (len(args) == 1 and args[0] == "-gui" and len(kwargs) == 0):
        call_gui()

    # To load a bunch of regions from files listed in a text file provide flag '-gui' and the filename(s)
    elif "-gui" in args and "filenames" in kwargs:
        instruction_lines = []
        loader_instructions = {
            "FP": [],
            "FT": [],
            "PE": [],
            "ES": [],
            "NC": [],
            "CO": [],
            "CROP": [],
            "CBG": [],
            "SBG": []
        }

        if "sections" in kwargs:
            sections_to_load = []
            sections_to_load += [sec.strip() for sec in kwargs["sections"].split(";")]
            if "filenames" in kwargs:
                for filename in kwargs["filenames"].split(';'):
                    instruction_lines += parse_batch_file(filename.strip(), sections_to_load)
        else:
            for filename in kwargs["filenames"].split(';'):
                instruction_lines += parse_batch_file(filename.strip())

        loaded_files = []
        if len(instruction_lines) > 0:
            for line in instruction_lines:
                if line is not None:
                    parts = [l.strip().split('=') for l in line.split(';')]
                    fname = parts[0][1].strip() # The name of the file to load is the first parameter in the line
                    # In case the file has already been put in the loading que, skip it
                    if fname in loaded_files:
                        continue
                    else:
                        loaded_files.append(fname)
                    for part in parts:
                        name, value = part[0].strip(), part[1].strip()
                        if name in loader_instructions:
                            if name == "NC" and value == '':
                                value = "1"
                            if name == "CROP" and value != '':
                                value = [v.strip() for v in value.split(':')]
                            if name == "CROP" and value == '':
                                value = [0, 0]  # To emphasize  that no cropping shall be done
                            loader_instructions[name].append(value)
        if len(loader_instructions["FP"]) > 0:
            call_gui("-batchload", **loader_instructions)
        else:
            print("No source files were loaded. Specqp process is terminated.")
            sys.exit()


if __name__ == "__main__":
    initialize_logging()
    # Setting up service variables, folders, log files path, etc.
    service.prepare_startup()
    specqp_logger.info(f"App started as: {sys.argv}")
    # Enable the case below to force the user to provide arguments
    # if len(sys.argv) < 2:
    #     raise SyntaxError("Insufficient arguments.")

    if len(sys.argv) == 1:
        # If only script name is specified call GUI for interactive work
        specqp_logger.info("Running the app in GUI mode")
        main()
    if len(sys.argv) > 1:
        # If there are additional arguments run in batch mode
        specqp_logger.info("Running the app in BATCH mode")
        args = []
        kwargs = {}
        for arg in sys.argv[1:]:
            if "=" in arg:
                key, val = arg.split('=')
                # If keyword arguments are provided
                kwargs[key.strip()] = val.strip()
            else:
                args.append(arg)
        main(*args, **kwargs)



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
      name='specqp',
      version='1.1.2',
      description='Quick plotting, correcting and fitting of spectroscopic data',
      long_description=long_description,
      long_description_content_type="text/markdown",
      url='https://github.com/Shipilin/specqp.git',
      author='Mikhail Shipilin',
      author_email='mikhail.shipilin@gmail.com',
      license='MIT',
      packages=find_packages(),
      include_package_data=True,
      unit_test='pytest',
      install_requires=['numpy', 'scipy', 'pandas', 'matplotlib', 'lmfit'],
      classifiers=["Programming Language :: Python :: 3",
                   "License :: OSI Approved :: MIT License",
                   "Intended Audience :: Science/Research",
                   "Natural Language :: English",
                   "Operating System :: MacOS :: MacOS X",
                   "Operating System :: Microsoft :: Windows",
                   "Environment :: MacOS X",
                   "Development Status :: 5 - Production/Stable"]
      )
